"""
File Purpose: tools whose main purpose is to work with implementation details of Python.

E.g. manipulate function docstrings
"""
import inspect
import string
import textwrap
import pydoc
import IPython  # for display

from .docs_tools import format_docstring


def printsource(obj, *, module=True, as_str=False):
    '''prints source code for object (e.g. call this on a function or class).
    module: bool
        whether to also tell module & type info about this object.
    as_str: bool
        if True, return result as a string instead of printing it.
    '''
    result = ''
    if module:
        if inspect.ismodule(obj):
            topline = f'module {obj.__name__}'
        else:
            mod = inspect.getmodule(obj)
            if mod is None:
                raise TypeError(f'inspect.getmodule(obj) is None, for obj with type={type(obj)}')
            else:
                mod = mod.__name__
            topline = f'{type(obj).__name__} {obj.__name__!r} from module {mod}'
        buffer = '-' * len(topline)  # line like "--------" of length len(topline)
        result += '# ' + topline + '\n'
        result += '# ' + buffer + '\n'
    result += str(inspect.getsource(obj))
    if as_str:
        return result
    else:
        print(result)

def displaysource(obj, *, module=True):
    '''display sourcecode for obj, including syntax highlighting. See also: printsource.
    module: bool
        whether to also tell module & type info about this object.
    '''
    source = printsource(obj, module=module, as_str=True)
    IPython.display.display(IPython.display.Code(source, language='python'))

def is_iterable(obj):
    '''returns whether obj is iterable. For super fast code, copy-paste instead of calling this function.'''
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True

def inputs_as_dict(callable_, *args, **kw):
    '''returns dict of all inputs to callable_ based on its signature and args & kwargs.
    raises TypeError if inputs would be invalid for callable_.
    Example:
        def foo(a, b=2, c=3, * d=4, e=5): pass
        inputs_as_dict(foo, 9, d=7, c=8) gives {'a':9, 'b':2, 'c':8, 'd':7, 'e':5}
        inputs_as_dict(foo, z=6) raises TypeError since foo doesn't accept kwarg 'z'.
    '''
    _iad_for_callable = _inputs_as_dict__maker(callable_)
    return _iad_for_callable(*args, **kw)

def _inputs_as_dict__maker(callable_):
    '''returns a function which returns dict of all inputs all inputs to callable_.'''
    f_signature = inspect.signature(callable_)
    def _inputs_as_dict(*args, **kw):
        '''returns dict of inputs as they would be named inside callable_'s namespace.
        includes params not input directly here, but defined by default for callable_.
        '''
        bound_args = f_signature.bind(*args, **kw)  # << will raise TypeError if inputs invalid for f.
        bound_args.apply_defaults()  # << include defaults
        params_now = bound_args.arguments  # << dict of {input name: value}.
        return params_now
    return _inputs_as_dict

def value_from_aliases(*aliaslist, **aliasdict):
    '''returns the single provided value from aliases.
    raises InputMissingError if all values are None.
    raises InputConflictError if values are not None and not the same (compared via 'is').
    '''
    found = False
    aliases = [*aliaslist, *aliasdict.values()]
    for a in aliases:
        if a is not None:
            if found and (value is not a):
                errmsg = f'Found multiple non-None values: args={aliaslist}, kwargs={aliasdict}'
                raise InputConflictError(errmsg)
            else:
                found = True
                value = a
    if not found:
        errmsg = f'Found no non-None values: args={aliaslist}, kwargs={aliasdict}'
        raise InputMissingError(errmsg)
    return value


_help_str_paramdocs = {
    'module': '''module: bool
        whether to include line with module name in help string.
        e.g. "function help_str in module PlasmaCalcs.tools.pytools"''',
    'blankline': '''blankline: bool
        whether to include blank line between module and signature.
        only considered if module=True AND signature=True.''',
    'signature': '''signature: bool
        whether to include line with signature in help string.
        e.g. "help_str(f, *, module=True, signature=True, indent=None)"''',
    'doc': '''doc: bool
        whether to include lines with docstring in help string.
        e.g. "return str for help(f)." ... and all the other docs in here.''',
    'indent': '''indent: str
        indent all lines in the result by this string.
        equivalent to textwrap.indent(result, indent).''',
}

@format_docstring(**_help_str_paramdocs)
def help_str(f, *, module=True, blankline=True, signature=True, doc=True, indent=''):
    '''return str for help(f).
    {module}
    {blankline}
    {signature}
    {doc}
    {indent}
    '''
    help0 = pydoc.render_doc(f, '%s', renderer=pydoc.plaintext)  # matches help(f)
    if module and blankline and signature and doc:  # including all parts of help0; no need to parse.
        return textwrap.indent(help0, indent)
    # otherwise, we need to parse help0 and remove some lines.
    lines = help0.splitlines()
    mod_lines = lines[0:1]  # line 0 is module name; line 1 is blank
    sig_lines = lines[2:3]  # line 2 is signature
    doc_lines = lines[3:]   # all remaining lines
    result_lines = []
    if module: result_lines.extend(mod_lines)
    if blankline and module and signature: result_lines.append('')
    if signature: result_lines.extend(sig_lines)
    if doc: result_lines.extend(doc_lines)
    result = '\n'.join(result_lines)
    result = textwrap.indent(result, indent)
    return result

def print_help_str(f, **kw_help_str):
    '''print help_str(f, **kw_help_str).'''
    print(help_str(f, **kw_help_str))

def indent_doclines(s, indent=' '*4):
    '''return string s, prepending all lines except line 0 with indent'''
    lines = s.splitlines()
    lines[1:] = [indent + line for line in lines[1:]]
    return '\n'.join(lines)

def indent_paramdocs(d, indent=' '*4):
    '''return dict but with indent_doclines called on all the values.'''
    return {key: indent_doclines(value, indent) for key, value in d.items()}


''' --------------------- String formatting --------------------- '''

def pad_missing_format_keys(s, keys):
    '''return s (a format string) with all missing keys padded with extra {}.
    E.g. ('{present} and {missing}', ['present']) --> '{present} and {{missing}}'

    keys: iterable of keys. (dict is also acceptable.)
    '''
    formatter = string.Formatter()
    result = ''
    for literal_text, field_name, format_spec, conversion in formatter.parse(s):
        result += literal_text
        if field_name is not None:
            field_content = '{' + field_name
            if conversion is not None:
                field_content += '!' + conversion
            if format_spec is not None:
                field_content += ':' + format_spec
            field_content += '}'
            if field_name not in keys:
                field_content = '{' + field_content + '}'
            result += field_content
    return result

def format_except_missing(s, **kw):
    '''return s.format(**kw) but ignoring any missing keys.'''
    s = pad_missing_format_keys(s, kw.keys())
    return s.format(**kw)

def replace_missing_format_keys(s, keys, replacement):
    '''return s (a format string) with all missing keys replaced by replacement.
    E.g. ('{present} and {missing}', ['present'], 'MISSING') --> 'present and MISSING'

    keys: iterable of keys. (dict is also acceptable.)
    '''
    formatter = string.Formatter()
    result = ''
    for literal_text, field_name, format_spec, conversion in formatter.parse(s):
        result += literal_text
        if field_name is not None:
            field_content = '{' + field_name
            if conversion is not None:
                field_content += '!' + conversion
            if format_spec is not None:
                field_content += ':' + format_spec
            field_content += '}'
            if field_name not in keys:
                field_content = replacement
            result += field_content
    return result

def format_replace_missing(s, replacement, **kw):
    '''return s.format(**kw) but replace any missing keys with replacement.
    E.g. ('{present} and {missing}', 'MISSING', present='yes') --> 'yes and MISSING'
    '''
    s = replace_missing_format_keys(s, kw.keys(), replacement)
    return s.format(**kw)



"""
File Purpose: loading global (i.e., fluid-independent) eppic.i inputs,
from values in an EppicInstabilityCalculator.
"""
import numpy as np
import xarray as xr

from ....errors import LoadingNotImplementedError
from ....other_calculators import InstabilityCalculator
from ....quantities import QuantityLoader
from ....tools import (
    simple_property, format_docstring,
    xarray_copy_kw, xarray_max, xarray_minimum,
)
from ....defaults import DEFAULTS


class EppicGlobInputsLoader(QuantityLoader):
    '''loads global (fluid-independent) eppic.i inputs, from values in an EppicInstabilityCalculator.'''

    # # # GLOB_INPUTS -- "DISPATCH" # # #

    # {varname: name for eppic.i file.}
    # with {x} indicating "replace with component (x, y, or z)",
    # name can be dict --> expect varname to expand into vars upon loading.
    # order here will be same order as used by eppic.i.
    GLOB_INPUTS = {
        # grid size in space
        'ndim_space': 'ndim_space',
        'safe_pow2_nspace': {
            'safe_nsubdomains': 'nsubdomains',
            'safe_pow2_nspace': 'n{x}',
        },
        'safe_dspace': 'd{x}',
        # grid size in time
        'safe_dt': 'dt',
        'safe_rounded_nt': 'nt',
        'safe_rounded_nout': 'nout',
        # physical parameters
        'm_neutral': 'm_neutral',
        'vth_neutral': 'vth_neutral',
        'B': 'B{x}',
        'E_un0': 'E{x}0_external',
        'fwidth': 'fwidth',  # (not physical, but related to E.)
        'fsteep': 'fsteep',  # (not physical, but related to E.)
        'eps0': 'eps',  # eps sets the units.
        # output-related vars (user probably never needs to touch these)
        'safe_iwrite': 'iwrite',
        'iread': 'iread',
        'hdf_output_arrays': 'hdf_output_arrays',
        'nout_avg': 'nout_avg',
        'npout': 'npout',
        'divj_out_subcycle': 'divj_out_subcycle',
    }

    @known_var(deps=GLOB_INPUTS)
    def get_glob_inputs(self):
        '''Dataset of global input values which got to eppic.i, here using PlasmaCalcs datavar names.
        Named & dimensioned here with PlasmaCalcs conventions,
            e.g. uses 'E_un0' which varies across 'component' dim,
            instead of labeling Ex0_external, Ey0_external, Ez0_external like in eppic.i.
        internally uses self.dspace_mode = self.safe_dspace_mode.
        Result data_vars names are the keys of self.GLOB_INPUTS.

        see also: eppici_glob, which uses eppic.i datavar names instead.
        '''
        with self.using(_cached_safe_dt=self('safe_dt')):  # [EFF] caching improves efficiency.
            return self(list(self.GLOB_INPUTS), dspace_mode=self.safe_dspace_mode)


    # # # SPATIAL SCALE: dx (and y and z) # # #

    cls_behavior_attrs.register('dspace_safety', default=DEFAULTS.EPPIC.DSPACE_SAFETY)
    dspace_safety = simple_property('_dspace_safety', setdefault=lambda: DEFAULTS.EPPIC.DSPACE_SAFETY,
        doc=f'''safety factor for safe dspace calculations. Larger is LESS safe.
        safe_dspace = unsafe_dspace * dspace_safety.
        Default: DEFAULTS.EPPIC.DSPACE_SAFETY (default: {DEFAULTS.EPPIC.DSPACE_SAFETY})''')

    @known_var
    def get_dspace_safety(self):
        '''safety factor for safe dspace calculations. Larger is LESS safe.
        safe_dspace = unsafe_dspace * dspace_safety.
        internally, value is stored (and can be adjusted) at self.dspace_safety.
        '''
        return xr.DataArray(self.dspace_safety)

    def _default_dspace_mode(self):
        '''default dspace_mode: 'safe_eqperp_ldebye_total'.'''
        return 'safe_eqperp_ldebye_total'

    @property
    def safe_dspace_mode(self):
        '''self.dspace_mode with 'safe_' prepended if self.dspace_mode doesn't already start with 'safe_'.'''
        mode = self.dspace_mode
        return mode if mode.startswith('safe_') else f'safe_{mode}'

    @property
    def unsafe_dspace_mode(self):
        '''self.dspace_mode with leading 'safe_' removed if self.dspace_mode starts with 'safe_'.'''
        mode = self.dspace_mode
        return mode[len('safe_'):] if mode.startswith('safe_') else mode

    DSPACE_MODE_OPTIONS = InstabilityCalculator.DSPACE_MODE_OPTIONS.copy()
    DSPACE_MODE_OPTIONS.update({f'safe_{k}': f'{k} * self.dspace_safety' for k in DSPACE_MODE_OPTIONS})
    _DSPACE_MODE_TO_DEPS = InstabilityCalculator._DSPACE_MODE_TO_DEPS.copy()
    _DSPACE_MODE_TO_DEPS.update({f'safe_{k}': ['dspace_safety'] + v for k, v in _DSPACE_MODE_TO_DEPS.items()})
    @known_var(attr_deps=[('dspace_mode', '_DSPACE_MODE_TO_DEPS')], aliases=['dspace'])
    def get_ds(self):
        '''spatial scale. (required by self('timescales') for speed-based timescales.)
        result depends on self.dspace_mode. See self.DSPACE_MODE_OPTIONS for options.
        '''
        mode = self.dspace_mode
        if mode.startswith('safe_'):
            safe_mode = mode
            mode = mode[len('safe_'):]
            safety = self('dspace_safety')
            with self.using(dspace_mode=mode):
                result = super().get_ds() * safety
            return result.assign_attrs(dspace_mode=safe_mode)
        else:
            return super().get_ds()

    @known_var(deps=['dspace', 'dspace_safety'])
    def get_safe_dspace(self):
        '''ds to use for simulation, including safety factor(s).
        Result depends on self.dspace_mode. See self.DSPACE_MODE_OPTIONS for options.
        Equivalent to self('dspace', dspace_mode=self.safe_dspace_mode).
        '''
        return self('dspace', dspace_mode=self.safe_dspace_mode)

    @known_var(deps=['dspace'])
    def get_unsafe_dspace(self):
        '''ds to use for simulation, without dspace_safety factor.
        Equivalent to self('dspace', dspace_mode=self.unsafe_dspace_mode).
        '''
        return self('dspace', dspace_mode=self.unsafe_dspace_mode)

    @known_var(deps=['safe_dspace'])
    def get_ds_for_timescales(self):
        '''spatial scale used when calculating timescales. Might be a vector, e.g. [dx, dy, dz].
        The method here returns self('safe_dspace').
        '''
        return self('safe_dspace')

    @known_var(deps=['safe_dspace', 'unsafe_dspace'])
    def get_direct_dspace_safety(self):
        '''direct_dspace_safety = safe_dspace / unsafe_dspace.'''
        return self('safe_dspace') / self('unsafe_dspace')


    # # # SPATIAL EXTENT: nx (and y and z) # # #

    cls_behavior_attrs.register('nspace_safety', default=DEFAULTS.EPPIC.NSPACE_SAFETY)
    nspace_safety = simple_property('_nspace_safety', setdefault=lambda: DEFAULTS.EPPIC.NSPACE_SAFETY,
        doc=f'''safety factor for safe Nspace calculations. Larger is MORE safe.
        safe_Nspace = unsafe_Nspace * nspace_safety.
        Default: DEFAULTS.EPPIC.NSPACE_SAFETY (default: {DEFAULTS.EPPIC.NSPACE_SAFETY})''')

    @known_var
    def get_nspace_safety(self):
        '''safety factor for safe Nspace calculations. Larger is MORE safe.
        safe_Nspace = unsafe_Nspace * nspace_safety.
        internally, value is stored (and can be adjusted) at self.nspace_safety.
        '''
        return xr.DataArray(self.nspace_safety)

    @known_var(deps=['spacescale_growth', 'safe_dspace'])
    def get_unsafe_Nspace(self):
        '''Nspace to use for simulation, without nspace_safety factor.
        unsafe_Nspace = spacescale_growth / safe_dspace.
        '''
        return self('spacescale_growth') / self('safe_dspace')

    @known_var(deps=['unsafe_Nspace', 'nspace_safety'])
    def get_safe_Nspace(self):
        '''number of grid cells (along each spatial dimension) to use for simulation,
        including safety factors(s).
        safe_Nspace = (spacescale_growth / safe_dspace) * nspace_safety
        '''
        return self('unsafe_Nspace') * self('nspace_safety')

    @known_var(deps=['safe_Nspace'])
    def get_safe_pow2_Nspace(self):
        '''number of grid cells (along each spatial dimension) to use for simulation,
        including safety factors, and ensuring result is a power of 2.
        (uses the smallest power of 2 such that result is still larger than safe_Nspace.)
        '''
        Nspace = self('safe_Nspace')
        return 2 ** np.ceil(np.log2(Nspace)).astype('int')

    @known_var(deps=['safe_pow2_Nspace', 'unsafe_Nspace'])
    def get_direct_nspace_safety(self):
        '''direct_nspace_safety = safe_pow2_Nspace / unsafe_Nspace.'''
        return self('safe_pow2_Nspace') / self('unsafe_Nspace')

    cls_behavior_attrs.register('nx_min', default=DEFAULTS.EPPIC.NX_MIN)
    nx_min = simple_property('_nx_min', setdefault=lambda: DEFAULTS.EPPIC.NX_MIN,
        doc=f'''minimum nx for input deck. Nx = nx * nsubdomains. See also: self.nsubdomains_max.
        Default: DEFAULTS.EPPIC.NX_MIN (default: {DEFAULTS.EPPIC.NX_MIN})''')

    cls_behavior_attrs.register('nsubdomains_max', default=DEFAULTS.EPPIC.NSUBDOMAINS_MAX)
    nsubdomains_max = simple_property('_nsubdomains_max', setdefault=lambda: DEFAULTS.EPPIC.NSUBDOMAINS_MAX,
        doc=f'''maximum nsubdomains for input deck. Nx = nx * nsubdomains. See also: self.nx_min.
        Default: DEFAULTS.EPPIC.NSUBDOMAINS_MAX (default: {DEFAULTS.EPPIC.NSUBDOMAINS_MAX})''')

    @known_var(deps=['safe_pow2_Nspace'])
    def get_safe_pow2_nspace(self):
        '''nx, ny, nz to use for simulation, including safety factors, and as a power of 2.
        result is a dataset of 'safe_pow2_nspace' and 'safe_nsubdomains'. nx is safe_Nx_sim/nsubdomains.
        result depends on self.nx_min and self.nsubdomains_max.
        if self('ndim_space')==2, also sets nz=1 everywhere.
        '''
        Nspace = self('safe_pow2_Nspace')
        if 'component' in Nspace.dims:
            raise LoadingNotImplementedError('[TODO] nspace from Nspace varying with component.')
        Nx = Nspace.assign_coords(component=self.components.get('x'))
        ny = Nspace.assign_coords(component=self.components.get('y'))
        nz = Nspace.assign_coords(component=self.components.get('z'))
        # break Nx into nx and nsubdomains...
        nx = self.nx_min
        nsubdomains = Nx // nx  # [TODO] assert nx evenly divides Nx.
        nsubdomains = np.minimum(nsubdomains, self.nsubdomains_max)
        nx = Nx // nsubdomains
        # if ndim_space==2, make nz==1...
        if self('ndim_space') == 2:
            nz = nz.copy(data=np.ones(nz.shape, dtype='int'))
        # make result:
        nspace = xr.concat([nx, ny, nz], dim='component')
        copy_kw = xarray_copy_kw(Nspace, array_to_dataset=True)
        return xr.Dataset({'safe_pow2_nspace': nspace, 'safe_nsubdomains': nsubdomains}, **copy_kw)

    @known_var(deps=['safe_pow2_nspace'])
    def get_safe_nsubdomains(self):
        '''nsubdomains to use for simulation. Nx = nx * nsubdomains.
        result depends on self.nx_min and self.nsubdomains_max.
        Also depends on self('safe_pow2_Nspace').
        '''
        return self('safe_pow2_nspace')['safe_nsubdomains']


    # # # TIME SCALE AND EXTENT: dt, nt # # #

    cls_behavior_attrs.register('dt_safety', default=DEFAULTS.EPPIC.DT_SAFETY)
    dt_safety = simple_property('_dt_safety', setdefault=lambda: DEFAULTS.EPPIC.DT_SAFETY,
        doc=f'''safety factor for safe dt calculations. Larger is LESS safe.
        safe_dt = unsafe_dt * dt_safety.
        Default: DEFAULTS.EPPIC.DT_SAFETY (default: {DEFAULTS.EPPIC.DT_SAFETY})''')

    @known_var
    def get_dt_safety(self):
        '''safety factor for safe dt calculations. Larger is LESS safe.
        safe_dt = unsafe_dt * dt_safety.
        internally, value is stored (and can be adjusted) at self.dt_safety.
        '''
        return xr.DataArray(self.dt_safety)

    @known_var(deps=['minf_timescale'])
    def get_unsafe_dt(self):
        '''dt to use for simulation, without dt_safety factor.
        unsafe_dt == minf_timescale.
        '''
        return self('minf_timescale')

    @known_var(deps=['dt_safety', 'unsafe_dt'])
    def get_safe_dt(self):
        '''dt to use for simulation, including safety factors: minf_timescale * dt_safety.
        result depends on self.dt_safety, as well as self.dspace_mode (for speed-based timescales).
        '''
        if hasattr(self, '_cached_safe_dt'):  # [EFF] if internally repeating safe_dt many times,
            return self._cached_safe_dt       # can do "with self.using(_cached_safe_dt=...):"
        return self('unsafe_dt') * self('dt_safety')

    @known_var(deps=['safe_dt', 'unsafe_dt'])
    def get_direct_dt_safety(self):
        '''safe_dt / unsafe_dt.'''
        return self('safe_dt') / self('unsafe_dt')

    cls_behavior_attrs.register('ntime_safety', default=DEFAULTS.EPPIC.NTIME_SAFETY)
    ntime_safety = simple_property('_ntime_safety', setdefault=lambda: DEFAULTS.EPPIC.NTIME_SAFETY,
        doc=f'''safety factor for safe nt calculations. Larger is MORE safe.
        safe_nt = unsafe_nt * ntime_safety.
        Default: DEFAULTS.EPPIC.NTIME_SAFETY (default: {DEFAULTS.EPPIC.NTIME_SAFETY})''')

    @known_var
    def get_ntime_safety(self):
        '''safety factor for safe nt calculations. Larger is MORE safe.
        safe_nt = unsafe_nt * ntime_safety.
        internally, value is stored (and can be adjusted) at self.ntime_safety.
        '''
        return xr.DataArray(self.ntime_safety)

    @known_var(deps=['timescale_growth'])
    def get_unsafe_nt(self):
        '''nt to use for simulation, excluding ntime_safety factor.
        unsafe_nt = timescale_growth / safe_dt
        '''
        return self('timescale_growth') / self('safe_dt')

    @known_var(deps=['unsafe_nt', 'ntime_safety'], aliases=['safe_ntime'])
    def get_safe_nt(self):
        '''nt to use for simulation, including safety factors.
        safe_nt = (timescale_growth / safe_dt) * ntime_safety.
        result is not an int; for int see self('safe_rounded_nt')
        '''
        return self('unsafe_nt') * self('ntime_safety')

    @known_var(deps=['safe_nt', 'safe_rounded_nout'])
    def get_safe_rounded_nt(self):
        '''nt to use for simulation, including safety factors and rounding.
        Rounded up to nearest multiple of self('safe_rounded_nout').
        '''
        ntime = self('safe_nt')
        multiple = self('safe_rounded_nout')
        nmul = np.ceil(ntime / multiple)
        return (nmul * multiple).astype('int')

    @known_var(deps=['safe_rounded_nt', 'unsafe_nt'])
    def get_direct_ntime_safety(self):
        '''direct_ntime_safety = safe_rounded_nt / unsafe_nt.'''
        return self('safe_rounded_nt') / self('unsafe_nt')


    # # # SNAP SAVE FREQUENCY: nout # # #

    cls_behavior_attrs.register('nout_waveprop_safety', default=DEFAULTS.EPPIC.NOUT_WAVEPROP_SAFETY)
    nout_waveprop_safety = simple_property('_nout_waveprop_safety',
        setdefault=lambda: DEFAULTS.EPPIC.NOUT_WAVEPROP_SAFETY,
        doc=f'''safety factor for safe nout calculations. Larger is LESS safe.
        safe_nout = min(safe_nout_waveprop, safe_nout_growth);
        safe_nout_waveprop = unsafe_nout_waveprop * nout_waveprop_safety.
        Default: DEFAULTS.EPPIC.NOUT_WAVEPROP_SAFETY (default: {DEFAULTS.EPPIC.NOUT_WAVEPROP_SAFETY})''')

    cls_behavior_attrs.register('nout_growth_safety', default=DEFAULTS.EPPIC.NOUT_GROWTH_SAFETY)
    nout_growth_safety = simple_property('_nout_growth_safety',
        setdefault=lambda: DEFAULTS.EPPIC.NOUT_GROWTH_SAFETY,
        doc=f'''safety factor for safe nout calculations. Larger is LESS safe.
        safe_nout = min(safe_nout_waveprop, safe_nout_growth);
        safe_nout_growth = unsafe_nout_growth * nout_growth_safety.
        Default: DEFAULTS.EPPIC.NOUT_GROWTH_SAFETY (default: {DEFAULTS.EPPIC.NOUT_GROWTH_SAFETY})''')

    @known_var
    def get_nout_waveprop_safety(self):
        '''safety factor for safe_nout_waveprop. Larger is LESS safe.
        safe_nout = unsafe_nout * nout_safety.
        internally, value is stored (and can be adjusted) at self.nout_waveprop_safety.
        '''
        return xr.DataArray(self.nout_waveprop_safety)

    @known_var
    def get_nout_growth_safety(self):
        '''safety factor for safe_nout_growth. Larger is LESS safe.
        safe_nout = unsafe_nout * nout_safety.
        internally, value is stored (and can be adjusted) at self.nout_growth_safety.
        '''
        return xr.DataArray(self.nout_growth_safety)

    @known_var(deps=['nout_waveprop_safety', 'nout_growth_safety'])
    def get_nout_safety(self):
        '''dataset of safety factors for safe nout calculations. Larger is LESS safe.
        safe_nout = unsafe_nout * nout_safety.
        internally, values are at self.nout_waveprop_safety and self.nout_growth_safety.
        result has 'nout_waveprop_safety' and 'nout_growth_safety' data_vars.
        '''
        return self(['nout_waveprop_safety', 'nout_growth_safety'])

    @known_var(deps=['abs_real_omega_at_growmax'])
    def get_waveprop_time(self):
        '''time for wave to propagate 1 full wavelength, according to real(omega) at growmax.
        waveprop_time = 2*pi/|real(omega at growmax)|.
        '''
        return 2 * np.pi / self('abs_real_omega_at_growmax')

    @known_var(deps=['waveprop_time', 'safe_dt'])
    def get_unsafe_nout_waveprop(self):
        '''nout to resolve wave propagation, without nout_waveprop_safety factor.
        unsafe_nout_waveprop = waveprop_time / safe_dt.
        '''
        return self('waveprop_time') / self('safe_dt')

    @known_var(deps=['unsafe_nout_waveprop', 'nout_waveprop_safety'])
    def get_safe_nout_waveprop(self):
        '''safe number of timesteps between snapshots, to well-resolve wave propagation:
        safe_nout_waveprop = (waveprop_time / safe_dt) * nout_waveprop_safety.
        '''
        return self('unsafe_nout_waveprop') * self('nout_waveprop_safety')

    @known_var(deps=['timescale_growth', 'safe_dt'])
    def get_unsafe_nout_growth(self):
        '''nout to resolve growth rate, without nout_growth_safety factor.
        unsafe_nout_growth = timescale_growth / safe_dt.
        '''
        return self('timescale_growth') / self('safe_dt')

    @known_var(deps=['unsafe_nout_growth', 'nout_growth_safety'])
    def get_safe_nout_growth(self):
        '''safe number of timesteps between snapshots, to well-resolve growth rate:
        safe_nout_growth = (growth timescale / safe_dt) * nout_growth_safety.
        '''
        return self('unsafe_nout_growth') * self('nout_growth_safety')

    @known_var(deps=['safe_nout_waveprop', 'safe_nout_growth'])
    def get_safe_nout(self):
        '''safe number of timesteps between snapshots, to well-resolve physical processes:
            - wave motion (by a distance of 1 wavelength)
            - growth rate (1 e-folding of growth)
        safe_nout = min(safe_nout_waveprop, safe_nout_growth). (Rounding not handled here;
            see safe_rounded_nout for a rounded value appropriate for eppic.i files.)
        '''
        return xarray_minimum(self('safe_nout_waveprop'), self('safe_nout_growth'))

    cls_behavior_attrs.register('nout_multiple', default=DEFAULTS.EPPIC.NOUT_MULTIPLE)
    nout_multiple = simple_property('_nout_multiple', setdefault=lambda: DEFAULTS.EPPIC.NOUT_MULTIPLE,
        doc=f'''ensure safe_rounded_nout is a multiple of this value * max safe_pow2_subcycle.
        E.g., nout_multiple=5 causes nout to be a multiple of 10,
            assuming safe_pow2_subcycle > 1 for at least 1 specie
            (subcycle=2^N with N>0 --> subcycle * 5 will be divisible by 10).
        Use nout_multiple=1 to ignore this value and just use the max safe_pow2_subcycle requirement.
        Default: DEFAULTS.EPPIC.NOUT_MULTIPLE (default: {DEFAULTS.EPPIC.NOUT_MULTIPLE})''')

    @known_var(deps=['safe_nout'])
    @format_docstring(nout_multiple=DEFAULTS.EPPIC.NOUT_MULTIPLE)
    def get_safe_rounded_nout(self):
        '''number of timesteps between snapshots, suitable for use in eppic.i file.
        At most safe_nout, but rounds down to integer, and might round down further to ensure
        result is a multiple of self.nout_multiple * safe_pow2_subcycle for all fluids
            (eppic requires subcycling to evenly divide nout)
            (self.nout_multiple is set by user; default {nout_multiple}.
                Use self.nout_multiple=1 to ignore nout_multiple restriction.
                Either way, will always have safe_pow2_subcycle restriction.)
        '''
        nout = self('safe_nout')
        subcycle = xarray_max(self('safe_pow2_subcycle'), 'fluid')
        # ^ because all safe_pow2_subcycle vals are like 2^N, can just take the max :)
        multiple = self.nout_multiple * subcycle
        nmul = nout // multiple   # nmul = largest N such that N * multiple < safe_nout.
        return (nmul * multiple).astype('int')

    @known_var(deps=['safe_rounded_nout', 'unsafe_nout_waveprop'])
    def get_direct_nout_waveprop_safety(self):
        '''safe_rounded_nout / unsafe_nout_waveprop'''
        return self('safe_rounded_nout') / self('unsafe_nout_waveprop')

    @known_var(deps=['safe_rounded_nout', 'unsafe_nout_growth'])
    def get_direct_nout_growth_safety(self):
        '''safe_rounded_nout / unsafe_nout_growth'''
        return self('safe_rounded_nout') / self('unsafe_nout_growth')

    @known_var(deps=['direct_nout_waveprop_safety', 'direct_nout_growth_safety'])
    def get_direct_nout_safety(self):
        '''dataset containing direct_nout_waveprop_safety and direct_nout_growth_safety.'''
        with self.using(_cached_safe_dt=self('safe_dt')):  # [EFF] caching improves efficiency.
            return self(['direct_nout_waveprop_safety', 'direct_nout_growth_safety'])

    @known_var(deps=['safe_rounded_nout'])
    @format_docstring(default_iwrite_nsnap=DEFAULTS.EPPIC.IWRITE_NSNAP)
    def get_safe_iwrite(self):
        '''safe_iwrite = safe_rounded_nout * iwrite_nsnap.
        (someday, iwrite_nsnap might be directly in EPPIC, but it's not there yet.
            For now, default iwrite_nsnap = {default_iwrite_nsnap})
        '''
        return self('safe_rounded_nout') * self('iwrite_nsnap')



"""
File Purpose: basic tests for eppic hookup in PlasmaCalcs.
These tests require an eppic.i file, but don't require any simulation output.
"""
import os
import pytest

import numpy as np

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


''' --------------------- basics --------------------- '''

def get_eppic_calculator(eppic_i='test_eppic_basics.i', **kw_init):
    '''make an eppic calculator'''
    with pc.InDir(HERE):
        ec = pc.EppicCalculator.from_here(eppic_i, **kw_init)
    return ec

def test_make_eppic_calculator():
    '''ensure can make an eppic calculator.'''
    get_eppic_calculator()


''' --------------------- dimensions --------------------- '''

def test_dimensions_logic():
    '''ensure dimensions logic is working properly.
    Assumes eppic.i has at least 2 distributions in it.
    '''
    ec = get_eppic_calculator(dist_names={0: 'e'})
    assert len(ec.fluids) == len(ec.input_deck.dists)
    # DimensionValue compared to int or str
    ec.fluid = None  # use all fluids
    assert ec.fluid[0] == 0
    assert ec.fluid[0] == 'e'
    # DimensionValueList indexing
    ec.fluid = None
    assert ec.fluid[(0,1)] == [0,1]
    assert ec.fluid[[0,1]] == [0,1]
    assert ec.fluid[:2] == [0,1]
    assert ec.fluid[:2] == ['e',1]
    assert ec.fluid[[1,0]] == [1,0]
    assert ec.fluid[[0,0,1,1,0]] == [0,0,1,1,0]
    # different ways to set the value of a dimension:
    ec.fluid = None
    assert len(ec.fluid) == len(ec.fluids)
    ec.fluid = 1
    assert ec.fluid == ec.fluids[1]
    ec.fluid = 'e'
    assert ec.fluid == ec.fluids[0]
    ec.fluid = ec.fluids[1]
    assert ec.fluid == ec.fluids[1]
    ec.fluid = slice(None)
    assert len(ec.fluid) == len(ec.fluids)
    assert all(f1==f2 for f1, f2 in zip(ec.fluid, ec.fluids))
    ec.fluid = [0]
    assert len(ec.fluid) == 1
    assert all(f1==f2 for f1, f2 in zip(ec.fluid, ec.fluids))
    ec.fluid = pc.ELECTRON
    assert ec.fluid == ec.fluids[0]
    ec.fluid = pc.IONS
    assert len(ec.fluid) == 2  # there are 2 ions in this input deck.
    ec.fluid = pc.CHARGED
    assert len(ec.fluid) == 3  # 2 ions + 1 electron = 3 charged fluids total.
    with pytest.raises(pc.FluidKeyError):
        ec.fluid = pc.NEUTRAL  # no neutral fluid in ec.fluids
    ec.jfluid = pc.NEUTRAL  # just test to ensure this works :)
    # 'using' block:
    ec.fluid = 1
    with ec.using(fluid=0):
        assert ec.fluid == 0
    assert ec.fluid == 1  # value is restored after exiting the block, normally.
    with pytest.raises(Exception):
        with ec.using(fluid=0):
            assert ec.fluid == 0
            raise Exception('boom - leave ec.using block via crash.')
            assert False, 'should have left block via crash...'
        assert ec.fluid == 1  # value is restored after exiting the block, due to crash.
    # 'iter'
    ec.fluid = 0
    for f in ec.iter_fluid():  # iterate current values only.
        assert f == 0
    ec.fluid = [0,1]
    for i, f in enumerate(ec.iter_fluid()):
        assert f == i
    assert ec.fluid == [0,1]   # original value restored afterwards.
    ec.fluid = 0
    for i, f in enumerate(ec.iter_fluids()):  # iterate all values
        assert f == i
        if i == len(ec.fluids)-1:
            break
    else:  # didn't break
        raise Exception('iter_fluids() never reached reach last fluid.')
    assert ec.fluid == 0   # original value restored afterwards.
    # 'iter' - enumerate
    assert list(enumerate(ec.iter_fluid())) == list(ec.enumerate_fluid())
    assert list(enumerate(ec.iter_fluids())) == list(ec.enumerate_fluids())
    assert list(ec.iter_fluid(enumerate=True)) == list(ec.enumerate_fluid())
    assert list(ec.iter_fluids(enumerate=True)) == list(ec.enumerate_fluids())

def test_ds_from_input_deck():
    '''ensure ec('ds') gets the right values, accounting for nout_avg.'''
    # this test exists because the original ec('ds') used input_deck['dx'] instead,
    #    which fails to account for nout_avg.
    ec = get_eppic_calculator()
    assert ec('ds_x') == ec.input_deck.Dx == ec.input_deck.get_dspace('x')
    assert ec('ds_y') == ec.input_deck.Dy == ec.input_deck.get_dspace('x')
    assert ec('ds_z') == ec.input_deck.Dz == ec.input_deck.get_dspace('x')


''' --------------------- loading vars --------------------- '''

def test_unknown_var():
    '''ensure crash with DimensionError or QuantCalcError when trying to get an unknown_var.'''
    ec = get_eppic_calculator()
    with pytest.raises((pc.DimensionError, pc.QuantCalcError)):
        ec('this_var_is_definitely_not_known_by_plasma_calcs')

def test_match_var():
    '''test match_var framework.'''
    ec = get_eppic_calculator()
    assert isinstance(ec.match_var('q'), pc.MatchedVar)
    assert isinstance(ec.match_var('q/m'), pc.MatchedPattern)
    q_over_m = ec.match_var('q/m')
    assert list(q_over_m.groups()) == ['q', 'm']
    assert list(q_over_m.deps) == [0,1]  # depends on groups()[0] and [1].
    assert isinstance(ec.match_var('(q*mod_B)/(m*nusn)'), pc.MatchedPattern)  # more complicated var
    # loading dims
    assert set(ec.match_var_loading_dims('q')) == {'fluid'}
    assert set(ec.match_var_loading_dims('q/m')) == {'fluid'}
    assert set(ec.match_var_loading_dims('B')) == {'component'}
    assert set(ec.match_var_loading_dims('mod_B')) == set()
    assert set(ec.match_var_loading_dims('q*B')) == {'fluid', 'component'}
    assert set(ec.match_var_loading_dims('q*mod_B')) == {'fluid'}
    assert set(ec.match_var_loading_dims('skappa')) == {'fluid'}

def test_nodata_vars():
    '''ensure can get some vars (which don't require any simulation data).'''
    ec = get_eppic_calculator()
    ec('q')
    ec('m')
    ec('B')
    ec('nusn')
    # info about neutrals is fully encoded in input deck too:
    ec('m_neutral')
    assert np.all(ec('m_n') == ec('m_neutral'))  # <-- ensures that aliases work
    ec('T_neutral')
    ec('u_neutral')
    with pytest.raises(pc.FormulaMissingError):
        ec('n_neutral')  # neutral density not known for eppic so this should crash.

def test_nodata_calcs():
    '''ensure can do some calculations (which don't require any simulation data).'''
    ec = get_eppic_calculator()
    ec.fluid = None
    assert np.all(ec('q/q')==1)
    assert np.all(ec('m-m')==0)
    assert np.all(ec('B_z')==ec('B', component='z')==ec('B', component=2))
    # more complicated calcs:
    skappa = ec('skappa')  # q |B| / (m nusn), depends only on fluids.
    assert skappa.size == len(ec.fluids)
    assert np.allclose(skappa, ec('q')*ec('mod_B')/(ec('m')*ec('nusn')))
    assert np.allclose(skappa, ec('q*mod_B/(m*nusn)'))        # parenthesis works
    assert np.allclose(skappa, ec('q*mod_B/m/nusn'))          # dividing by multiple things also works.
    assert np.allclose(skappa, ec('(q*(mod_B))/(m*(nusn))'))  # adding extra parentheses is fine
    assert np.allclose(1, ec('skappa/(q*mod_B/(m*nusn))'))    # dividing by complicated var is fine

def test_set_vars_from_input():
    '''ensure can set vars from input deck.'''
    ec = get_eppic_calculator()
    ec.set_vars_from_inputs()

def test_get_vars_from_input():
    '''ensure can get vars after setting them from input deck.'''
    ec = get_eppic_calculator()
    with pytest.raises((pc.DimensionError, pc.QuantCalcError)):
        ec('n')  # can't get 'n' before setting vars from input
    with pytest.raises((pc.DimensionError, pc.QuantCalcError)):
        ec('u')  # can't get 'u' before setting vars from input
    with pytest.raises((pc.DimensionError, pc.QuantCalcError)):
        ec('T')  # can't get 'T' before setting vars from input
    ec.set_vars_from_inputs()
    n = ec('n')   # can get 'n' after setting vars from input
    u = ec('u')   # can get 'u' after setting vars from input
    T = ec('T')   # can get 'T' after setting vars from input
    # size test for n (T is similar; not tested here)
    assert ec('n').size == len(ec.fluids)
    assert np.all(ec('n') == [fluid['n0d'] for fluid in ec.iter_fluid()])
    assert np.all(ec('n').isel(fluid=[1,0]) == [ec.fluid[1]['n0d'], ec.fluid[0]['n0d']])
    with ec.using(fluid=0):
        assert ec('n') == ec.fluid['n0d']
    # size test for u
    with ec.using(fluid=None, component=None):
        assert ec('u').size == ec.current_n_fluid() * ec.current_n_component()
        assert ec('u').size == ec.current_n_dimpoints(['fluid', 'component'])
    with ec.using(fluid=[0,1], component='y'):
        assert ec('u').size == ec.current_n_dimpoints(['fluid', 'component'])
    with ec.using(fluid=[0,1], component=None):
        assert ec('u').size == ec.current_n_dimpoints(['fluid', 'component'])
    with ec.using(fluid=0, component=('x', 'y')):
        assert ec('u').size == ec.current_n_dimpoints(['fluid', 'component'])
    # calcs with u. Also Ta (Ta nontrivial; u sometimes trivial, i.e. all 0)
    assert np.all(ec('mod_u') == np.sqrt(ec('u_x')**2 + ec('u_y')**2 + ec('u_z')**2))
    assert np.all(ec('mod_Ta') == np.sqrt(ec('Ta_x')**2 + ec('Ta_y')**2 + ec('Ta_z')**2))


''' --------------------- misc vars or aspects of EppicCalculator --------------------- '''

def test_get_string_valued_var():
    '''ensure can get collision_type (a string-valued var).'''
    ec = get_eppic_calculator()
    arr = ec('collision_type', fluid=0, jfluid=0)
    assert arr.size == 1
    arr = ec('collision_type', fluid=None, jfluid=0)
    assert arr.size == len(ec.fluids)
    arr = ec('collision_type', fluid=None, jfluid=None)
    assert arr.size == len(ec.fluids) * len(ec.jfluids)
    # grid of collision types (jfluid was probably just an EppicNeutral before)
    ec.jfluids = ec.fluids
    ec.jfluid = None
    arr = ec('collision_type', fluid=None, jfluid=None)
    assert arr.size == len(ec.fluids) * len(ec.jfluids)
    assert arr[0,0] == '0'  # no collisions between fluid and itself.
    assert ec.fluids[0].is_electron()
    assert ec.fluids[1].is_ion()
    assert arr[0,1] == 'coulomb'  # electron-ion collisions are coulomb collisions
    assert ec.fluids[2].is_ion()
    assert arr[1,2] == 'coulomb'  # ion-ion collisions are coulomb collisions.



"""
File Purpose: labels, e.g. titles, suptitles
"""
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings, PlotSettingsMixin
from .plot_tools import infer_movie_dim

from ..errors import PlottingNotImplementedError, DimensionKeyError
from ..tools import (
    UNSET,
    alias_child, alias_key_of, simple_property,
    join_strs_with_max_line_len,
    xarray_nondim_coords, xarray_fill_coords,
    pcAccessor, xarray_isel, xarray_dims_coords,
)


_paramdocs_labels = {
    'txt': '''Text object
        e.g., the result of plt.text(...) or plt.title(...)''',
    'frame_to_text': '''list, dict, or callable.
        map from frame number to title string. title for frame will be:
        callable --> frame_to_text(frame)
        list or dict --> frame_to_text[frame]''',
    'array': '''xarray.DataArray, probably ndim=3.
        the full DataArray which will be plotted throughout the movie.
        internally, store xarray_fill_coords(array), so that coordless dims' indices can be used.
            e.g. if "dim0" is a dimension with no coords, will use np.arange(dim0.size).''',
    'base_text': '''UNSET or str
        the base text string, to be formatted by array at each frame.
        UNSET --> will use array._title_for_slice() for array at each frame.
        str --> will be formatted by xarray_nondim_coords(array at frame).
                E.g., base_text='fluid={{fluid}}, time={{t:.2e}} seconds'.''',
    't': '''None or str
        the array dimension which frames will index. E.g. 'time'.
        None --> infer it via infer_movie_dim(array.dims, t)''',
}

''' --------------------- TextPlotElement / MovieText / XarrayText --------------------- '''

@PlotSettings.format_docstring(**_paramdocs_labels)
class TextPlotElement(MoviePlotElement):
    '''MoviePlotElement for Text object. see MovieText for updating text over time.

    txt: {txt}
    '''
    def __init__(self, txt, **kw_super):
        super().__init__(**kw_super)
        if isinstance(txt, str):
            raise TypeError(f'Expected txt to be a matplotlib.text.Text object, not a string.')
        self.txt = txt
        self.text = txt.get_text()

    text = alias_key_of('data', 'text',
        doc='''current text string. Internally stored at self.data['text']''')
    txt = simple_property('_txt', default=None, doc='''matplotlib.text.Text object''')
    ax = alias_child('txt', 'axes', if_no_child=None, doc='''axes containing this Text.''')
    fig = alias_child('txt', 'figure', if_no_child=None, doc='''figure containing this Text.''')

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''update the plot using data['text'].
        return list of all updated matplotlib Artist objects, for FuncAnimation(..., blit=True),
            i.e. [] if self.ax is None else [self.txt].
            (FuncAnimation fails to blit any artists not attached to axes,
            and self.txt is attached to axes if and only if self.ax is not None)
        '''
        self.text = data['text']
        self.txt.set_text(self.text)
        return [] if self.ax is None else [self.txt]

    # # # DISPLAY # # #
    REPR_TEXT_MAX_LEN = 30  # text longer than this will be abbreviated in repr.

    def __repr__(self):
        text = self.text
        if isinstance(text, str):
            if len(text) > self.REPR_TEXT_MAX_LEN:
                text = text[:self.REPR_TEXT_MAX_LEN] + '...'
        return f'{type(self).__name__}({text!r})'


@PlotSettings.format_docstring(**_paramdocs_labels)
class MovieText(MoviePlotNode):
    '''MoviePlotNode for a Text object.

    txt: {txt}
    frame_to_text: {frame_to_text}
    '''
    element_cls = TextPlotElement

    def __init__(self, txt, frame_to_text, **kw_super):
        self.txt = txt
        self.frame_to_text = frame_to_text
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    ax = alias_child('obj', 'ax', doc='''mpl.axes.Axes where this MovieText is plotted.''')
    @ax.getter
    def ax(self):  # ax=None if not self.plotted
        return self.obj.ax if self.plotted else None

    fig = alias_child('obj', 'fig', doc='''figure where this MovieText is plotted.''')
    @fig.getter
    def fig(self):  # fig=None if not self.plotted
        return self.obj.fig if self.plotted else None

    text = property(lambda self: self.obj.text if self.plotted else None,
        doc='''the text string stored in self.obj; use self.update_to_frame to update it.''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the TitlePlotElement at self.obj.'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs('pc.TextPlotElement')
        # >> actually plot the text <<
        self.obj = self.element_cls(self.txt, **kw_plot)
        self.obj.update_data(data)
        # bookkeeping
        self.frame = frame

    def get_data_at_frame(self, frame):
        '''returns {'text': text string for this frame}.'''
        if callable(self.frame_to_text):
            text = self.frame_to_text(frame)
        else:
            text = self.frame_to_text[frame]
        return {'text': text}

    def get_nframes_here(self):
        '''returns number of frames that could be in this movie, based on this node.'''
        if callable(self.frame_to_text):
            errmsg = (f'{type(self).__name__}.frame_to_text is callable, and frames not provided.\n'
                      'Debugging: provide self.frames=int, or, provide list or dict frame_to_text.')
            raise PlottingNotImplementedError(errmsg)
        return len(self.frame_to_text)


@pcAccessor.register('text', totype='array')
@PlotSettings.format_docstring(**_paramdocs_labels)
class XarrayText(MovieText):
    '''MoviePlotNode managing text associated with an xarray.DataArray.

    txt: {txt}
    array: {array}
    t: {t}
    base_text: None, {base_text}
        if None, infer from txt.

    --- Examples ---
        # simple example:
        import PlasmaCalcs as pc
        array = ...  # some array having fluid & t coords...
        txt = plt.text(0, 7, '{{fluid}}, t={{t:.2e}}')  # text at x=0, y=7 in data coords.
        xtext = pc.XarrayText(array, txt, t='t')
        xtext.save('text_movie.mp4')

        # slightly more complex: start like above, but attach to existing MoviePlotNode instead:
        xim = array.pc.image()
        xim.add_child(xtext, arr)
        xim.save('image_and_text_movie.mp4')  # movie of image & text, both updating in time!

        # one more notable option: instead of txt = plt.text(...),
        # could do txt = plt.title(...), plt.suptitle(...) or similar option.
        # also, if infer_movie_dim can figure out t, don't need to provide it explicitly:
        xtext = pc.XarrayText(array, plt.title('{{fluid}}, t={{t:.2e}}'))

        # finally, note that this can be accessed from array, e.g.:
        xtext = array.pc.text(plt.title('{{fluid}}, t={{t:.2e}}'))
    '''
    def __init__(self, array, txt, t=None, *, base_text=None, **kw_super):
        # text bookkeeping
        self.txt = txt
        if isinstance(txt, str):
            errmsg = (f'Expected txt = matplotlib.text.Text object, not a string (got txt={txt!r})\n'
                      'Maybe you forgot to use txt=plt.text(...), txt=plt.title(...), or similar.')
            raise TypeError(errmsg)
        if base_text is None:
            base_text = txt.get_text()
        self.base_text = base_text
        # array bookkeeping
        self.array = xarray_fill_coords(array)
        try:
            self.t = infer_movie_dim(array.dims, t)
        except DimensionKeyError:
            if t in array.coords:
                self.t = t  # movie will only have 1 frame, but it's fine...
            else:
                raise
        # plot stuff
        super(MovieText, self).__init__(**kw_super)

    def frame_to_text(self, frame):
        '''return the text string at this frame'''
        if self.t is None:
            arr_at_frame = self.array
        else:
            arr_at_frame = xarray_isel(self.array, {self.t: frame})
        if self.base_text is UNSET:
            result = arr_at_frame._title_for_slice()
        else:
            vals = xarray_nondim_coords(arr_at_frame, scalars_only=True)
            result = self.base_text.format(**vals)
        return result

    def get_nframes_here(self):
        '''returns number of frames that could be in this movie, based on this node.
        0 if self never updates (i.e., self.t is None).
        '''
        return 0 if self.t is None else self.array.coords[self.t].size


''' --------------------- Titles & Suptitles --------------------- '''

@pcAccessor.register('title', totype='array')
@PlotSettings.format_docstring(**_paramdocs_labels)
def xarray_title_plot_node(array, base_text=UNSET, t=None, *, ax=None, parent=None, **kw):
    '''create a MoviePlotNode for a title associated with an xarray.DataArray.

    array: {array}
    base_text: {base_text}
    t: {t}

    parent: None or MoviePlotNode
        if provided, the parent of this node. None -> this node has no parent.

    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    text_kw: {text_kw}

    Additional kwargs go directly to plt.title()
    '''
    # title settings
    ps = PlotSettings(**kw, pop=True)
    kw_text = ps.get('text_kw', default=dict()).copy()
    kw_title = {**kw_text, **ps.get('title_kw', default=dict())}
    kw_title.setdefault('y', ps.get('title_y', default=None))
    fontfamily = ps.get('title_font')
    if fontfamily is not None: kw_title.update(fontfamily=fontfamily)
    # ax
    if ax is None:
        ax = plt.gca()
    # make title Text object
    txt = ax.set_title(str(base_text), **kw_title)
    # make & return MoviePlotNode
    # (use array.pc.text, not XarrayText, in case other module attaches XarrayText subclass.)
    result = array.pc.text(txt, t=t, base_text=base_text, parent=parent, **ps.kw)
    return result


@pcAccessor.register('suptitle', totype='array')
@PlotSettings.format_docstring(**_paramdocs_labels)
def xarray_suptitle_plot_node(array, base_text=UNSET, t=None, *, fig=None, parent=None, **kw):
    '''create a MoviePlotNode for a suptitle associated with an xarray.DataArray.

    array: {array}
    base_text: {base_text}
    t: {t}

    parent: None or MoviePlotNode
        if provided, the parent of this node. None -> this node has no parent.

    suptitle_font: {suptitle_font}
    suptitle_y: {suptitle_y}
    suptitle_kw: {suptitle_kw}
    text_kw: {text_kw}

    Additional kwargs go directly to plt.suptitle()
    '''
    # suptitle settings
    ps = PlotSettings(pop_from=kw)  # pop setting from kwargs
    kw_text = ps.get('text_kw', default=dict()).copy()
    kw_suptitle = {**kw_text, **ps.get('suptitle_kw', default=dict())}
    kw_suptitle.setdefault('y', ps.get('suptitle_y', default=None))
    fontfamily = ps.get('suptitle_font')
    if fontfamily is not None: kw_suptitle.update(fontfamily=fontfamily)
    # fig
    if fig is None:
        fig = plt.gcf()
    # make suptitle Text object
    txt = fig.suptitle(str(base_text), **kw_suptitle)
    # make & return MoviePlotNode
    # (use array.pc.text, not XarrayText, in case other module attaches XarrayText subclass.)
    result = array.pc.text(txt, t=t, base_text=base_text, parent=parent, **ps.kw)
    return result


''' --------------------- SUBPLOTS Titles & Suptitles --------------------- '''

def title_from_coords(array, cnames=None, *, width=None, formattable=False):
    '''return a decent title from scalar coords.
    cnames: None or list of str
        coords to include values of in the title.
        None --> include all scalar coords.
    width: None or int
        max number of characters per line, before putting a newline between coords.
    formattable: bool
        whether to return a formattable version of the title.
    '''
    if cnames is None:
        cnames = xarray_nondim_coords(array, scalars_only=True)
    # use pretty number formatting if possible, else just str().
    results = []  # strs for title, e.g. 'fluid={fluid:}' or 'z={z:.3g}'
    examples = dict()  # {result: str with example val plugged in}, e.g. 'fluid=e-' or 'z=3.14'
    for c in cnames:
        v_example = array.coords[c].values.flat[0]  # example value of this coord.
        if c.endswith('_index') and np.issubdtype(array.coords[c].dtype, np.integer):
            fstr = 'd'  # special rule for index: use full int
        else:
            fstr = '.3g'
        try:
            f'{v_example:{fstr}}'
        except Exception:  # pretty formatting failed (e.g. non-numeric data).
            fstr = ''
        # [TODO] use !r if str type
        tmp = f'{c}={{{c}:{fstr}}}'
        results.append(tmp)
        examples[tmp] = tmp.format(**{c: v_example})
    # join strs with commas and/or newlines while respecting max width.
    result = join_strs_with_max_line_len(results, sep=', ', max=width,
                                         key=lambda r: len(examples[r]))
    if formattable:
        return result
    else:
        return result.format(**{c: array.coords[c].item() for c in cnames})


@PlotSettings.format_docstring()
class XarraySubplotTitlesInferer(PlotSettingsMixin):
    '''infers base titles & suptitle to use for subplots.
    "Base" means before title.format(nondim coord values from the array)

    array: xarray.DataArray
        infer titles from this array, after filling any missing coords
            (e.g. if dim_1 has no coords, use np.range(dim_1.size))
    t: None or str
        dimension name for the time axis (for movies).
    row: None or str
        dimension to plot ACROSS rows.
        None --> not plotting anything across rows; ncols=1.
    col: None or str
        dimension to plot DOWN columns.
        None --> not plotting anything across columns; nrows=1.
    subplot_title_width: {subplot_title_width}
    suptitle_width: {suptitle_width}
    '''
    def __init__(self, array, t=None, *, row=None, col=None, **kw_super):
        self.array = xarray_fill_coords(array)
        self.t = t
        self.row = row
        self.col = col
        super().__init__(**kw_super)

    def title_from_coords(self, cnames=None, *, width=None, formattable=True):
        '''return a decent title from scalar coords.
        cnames: None or list of str
            coords to include values of in the title.
            None --> include all scalar coords.
        width: None or int
            max number of characters per line, before putting a newline between coords.
        formattable: bool
            whether to return a formattable version of the title.
        '''
        return title_from_coords(self.array, cnames=cnames, width=width, formattable=formattable)

    def infer_title(self):
        '''infers title to use for subplots.
        This will just include the row & col dims & any associated coords.
        (excluding coords which will still be multidimensional within each subplot.)
        (also excluding MultiIndex coords; they are ugly, and all their info
            is already in title, via the coords within the MultiIndex.)
        '''
        array = self.array
        dims_coords = xarray_dims_coords(array)
        cnames = set()  # names of coords to include in title
        if self.row is not None:
            cnames.update(dims_coords[self.row])
        if self.col is not None:
            cnames.update(dims_coords[self.col])
        # remove coords which will be multi-dimensional within each subplot,
        #  i.e. have at least 1 dim other than just row and/or col.
        multi_dim_coords = set()
        for c in cnames:
            if len(set(array[c].dims) - {self.row, self.col}) > 0:
                multi_dim_coords.add(c)
        cnames = cnames - multi_dim_coords
        # remove MultiIndex coords
        multi_index_coords = set()
        for (idx, _idict) in array.indexes.group_by_index():
            if isinstance(idx, pd.MultiIndex):
                multi_index_coords.add(idx.name)
        cnames = cnames - multi_index_coords
        # style formatting, then get result:
        cnames = [c for c in self.array.coords if c in cnames]  # style: same order as in array.
        result = self.title_from_coords(cnames,
                        width=self.plot_settings['subplot_title_width'])
        return result

    def infer_suptitle(self):
        r'''infers title to use for subplots.
        This may include the array name, units, t dim,
            and any scalar xarray_nondim_coords not associated with row or col.

        formatting will be nice depending on which pieces of info exist.
            name --> '{name} {units}\n{t}\n{nondim_coords}'
            no name --> '{t} {units}\n{nondim_coords}'
            no t --> '{nondim_coords} {units}'
            if no units info remove {units} part
            if no scalar nondim_coords_values remove \n{nondim_coords} part
        '''
        array = self.array
        name = array.name
        t = self.t
        infos = dict()
        # array.name info
        if name is not None:
            infos['name'] = name
        # t info
        if t is not None:
            t = str(t)
            if t in ('time', 't', 'x', 'y', 'z'):
                infos['t'] = 'dim = {dim:.2e}'.replace('dim', t)
            elif (t == 'snap') and ('t' in array.coords):
                infos['t'] = 'snap = {snap!s:6s}, t = {t:.2e}'
            else:
                infos['t'] = 'dim = {dim!s}'.replace('dim', t)
        # units info
        if 'units' in array.attrs:
            units = array.attrs['units']
            infos['u'] = f'[{units}]'
        # scalar nondim coords - use pretty number formatting if possible, else just str()
        ncoords = xarray_nondim_coords(array, scalars_only=True)
        if len(ncoords) > 0:
            infos['c'] = self.title_from_coords(ncoords,
                                width=self.plot_settings['suptitle_width'])
        # return result, formatted based on which pieces of info exist.
        if infos:
            if 'name' in infos:
                uinfo = f" {infos['u']}" if 'u' in infos else ''
                tinfo = f"\n{infos['t']}" if 't' in infos else ''
                cinfo = f"\n{infos['c']}" if 'c' in infos else ''
                return f"{infos['name']}{uinfo}{tinfo}{cinfo}"
            elif 't' in infos:
                uinfo = f" {infos['u']}" if 'u' in infos else ''
                cinfo = f"\n{infos['c']}" if 'c' in infos else ''
                return f"{infos['t']}{uinfo}{cinfo}"
            elif 'c' in infos:
                uinfo = f" {infos['u']}" if 'u' in infos else ''
                return f"{infos['c']}{uinfo}"
            elif 'u' in infos:
                return f"units={infos['u']}"
        return None  # no info to add.



"""
File Purpose: misc. tools for reading directly from Copapic files
"""
import collections
import datetime
import os
import re
import pandas as pd
import xarray as xr
import json

# from ...errors import (
#     ValueLoadingError,
#     FileContentsError, FileContentsConflictError, FileContentsMissingError,
#     InputError,
# )
from ...tools import (
    InDir
)

''' --------------------- read copapic.json file --------------------- '''

def read_copapic_json_file(filename):
    '''Parse Copapic input from the file, returning (dict of global vars, dict of dists).
    dict of global vars has {varname: value} for all pairs not inside a distribution.
    dict of dists has {N: {varname: value} for all pairs in dist} for each dist;
        for each varname in a dist, ensure varname ends with str(N) (append it if necessary).
        the N's will all be stored here as ints.

    Arguments:
    filename: string
        file to read from.
    '''
    deck = json.load(open(filename, 'r'))
    return deck


def read_copapic_snaps_info(snaps_from, dt, dir=os.curdir):
    '''returns (snapnames, times).
    snaps_from: string
        directory of output files. Corresponds to the title of the simulation.
    dir: string (default '.')
        directory of 'copapic.json' file.
    dt: None or value
        the simulation time step; each snap's time is equal to dt * (snap name)
        if None, attempt to read 'dt' from copapic.json file.
    '''
    with InDir(dir):  # temporarily cd to dir; restore original dir after.
        # get snaps
        snapnums = []
        if not os.path.isdir(snaps_from):
            return ([], [])
        for filename in os.listdir(snaps_from):  # directory named 'parallel'
            match = re.match(r'data_(\d+)\.', filename[:-2])
            if match:
                snapname, = match.groups()
                snapnums.append(int(snapname))

    snapnums = sorted(snapnums)
    snapnames, times = [], []
    for snapnum in snapnums:
        snapnames.append(str(snapnum))
        times.append(dt * snapnum)
    return (snapnames, times)


"""
File Purpose: math tools
"""

import ast
import functools
import math
import operator as op

import numpy as np

''' --------------------- ast math eval --------------------- '''

def ast_math_eval(expr):
    '''evaluate a string math expression using ast. (only accepts numbers & math operations.)
    This is safe because it uses ast; it does not ever use the unsafe eval() command.
    '''
    node = ast.parse(expr, mode='eval').body
    return _ast_math_eval_node(node)

# dict of {ast class: associated operator}
AST_OPS = {
        ast.Add: op.add,
        ast.Sub: op.sub,
        ast.Mult: op.mul,
        ast.Div: op.truediv,
        ast.FloorDiv: op.floordiv,
        ast.Pow: op.pow,
        ast.USub: op.neg,
        }

def _ast_math_eval_node(node):
    '''evaluate a parsed ast node, recursively. For internal use. See ast_math_eval() for user interface.'''
    if isinstance(node, ast.Constant):
        return node.value
    elif isinstance(node, ast.BinOp):
        this_op = AST_OPS[type(node.op)]
        return this_op(_ast_math_eval_node(node.left), _ast_math_eval_node(node.right))
    elif isinstance(node, ast.UnaryOp):
        this_op = AST_OPS[type(node.op)]
        return this_op(_ast_math_eval_node(node.operand))
    else:
        raise TypeError(node)


''' --------------------- rounding --------------------- '''

def round_to_int(x, mode='round'):
    '''return x rounded to an integer, using the specified mode.
    (if x is already an integer, returns x unchanged.)

    mode: 'round', 'floor', 'ceil' or 'int'
        'round' --> as per builtins.round(). round to nearest integer, ties toward even integers.
        'int' --> as per builtins.int(). always rounds towards 0.
        'floor' --> round towards negative infinity.
        'ceil' --> round towards positive infinity.
    '''
    if mode == 'round':
        return round(x)
    elif mode == 'int':
        return int(x)
    elif mode == 'floor':
        return math.floor(x)
    elif mode == 'ceil':
        return math.ceil(x)
    else:
        raise ValueError(f'unrecognized mode={mode!r}, expected "round", "int", "floor", or "ceil".')

def float_rounding(x, prec=8):
    '''round x to a more-likely-to-be-input float (if possible, else just return x).
    [TODO] change this to work even if a decimal point appears inside the 0's or 9's.

    Examples:
        0.20000000001, prec=8 --> float(0.2)
        35999999999.0, prec=8 --> float(36000000000)
        35999999999.0, prec=12 --> unchanged (because prec > number of consecutive 0's or 9's)
        0.12345678901234 --> unchanged (because no sequence of consecutive 0's or 9's)

    prec: int
        number of consecutive 0's or 9's required in str(x) before rounding.
    '''
    s = str(float(x))
    round0str = '0' * prec
    round9str = '9' * prec
    if round0str in s:
        i = s.index(round0str)
        result, remaining = s[:i], s[i:]
    elif round9str in s:
        i = s.index(round9str)
        if i == 0:
            return x  # no rounding needed, x just starts like 9999999...
        result, remaining = s[:i], s[i:]
        result = result[:-1] + str(int(result[-1]) + 1)  # add 1 to final digit; we are rounding it up.
    else:
        return x  # no obvious rounding to do.
    if '.' in remaining:  # we are before the decimal point.
        d = remaining.index('.')
        result += '0' * d   # append enough 0's to reach the decimal point.
        if 'e' in remaining:  # if e follows, append the exponent. (probably never used)
            e = remaining.index('e')
            result += remaining[e:]
    elif 'e' in remaining:  # we are before the exponent and also before the decimal point. (probably never used)
        e = remaining.index('e')
        result += '0' * e + remaining[e:] # append the appropriate number of 0's, then the exponent.
    return float(result)


''' --------------------- classify math objects --------------------- '''

def is_integer(x):
    '''return whether x is an integer.
    by first checking isinstance(x, int),
    if that's False, check x.is_integer() if it exists,
    else return False.
    '''
    if isinstance(x, int):
        return True
    else:
        try:
            return x.is_integer()
        except AttributeError:
            return False


''' --------------------- simple math --------------------- '''

def product(iterable):
    '''returns the product of all elements in iterable.
    if len(iterable) == 0, returns 1.
    However, if len(iterable) > 0, starts with the first element instead of 1.
    (This is useful when using non-numeric objects which have defined __mul__ method.)
    '''
    l = list(iterable)
    if len(l) == 0:
        return 1
    return nonempty_product(iterable)

def nonempty_product(iterable):
    '''returns the product of all elements in iterable; iterable must have at least 1 element.
    Crash with TypeError if iterable has no elements.
    Does not assume that 1 is the multiplicative identity.
        E.g. does l[0] * l[1] * ... * l[-1], where l = list(iterable).
    (This is useful when using non-numeric objects which have defined __mul__ method.)
    Equivalent to functools.reduce((lambda x,y: x*y), iterable).
    '''
    return functools.reduce(op.mul, iterable)


''' --------------------- numpy math --------------------- '''

def np_all_int(x):
    '''returns whether all values in x are integer-valued. (regardless of x's type).
    Equivalent: np.all(np.mod(x, 1) == 0)
    '''
    return np.all(np.equal(np.mod(x, 1), 0))



''' --------------------- roman numerals --------------------- '''

def as_roman_numeral(x):
    '''convert x to a roman numeral string.'''
    if not 0 < x < 4000:
        raise NotImplementedError(f'as_roman_numeral(x) with x <= 0 or x >= 4000. Got x={x}')
    if not is_integer(x):
        raise TypeError(f'as_roman_numeral(x) with non-integer x. Got x={x}')
    num_map = [(1000, 'M'), (900, 'CM'),
                (500, 'D'), (400, 'CD'),
                (100, 'C'), (90, 'XC'),
                (50, 'L'), (40, 'XL'),
                (10, 'X'), (9, 'IX'),
                (5, 'V'), (4, 'IV'),
                (1, 'I')]
    number = x
    result = ''
    for n, r in num_map:
        while number >= n:
            result = result + r
            number = number - n
    return result

def from_roman_numeral(s):
    '''convert roman numeral string s to an integer.'''
    if not isinstance(s, str):
        raise TypeError(f'from_roman_numeral(s) with non-string s. Got s={s}')
    if len(s) == 0:
        raise ValueError(f'from_roman_numeral(s) with empty string s.')
    num_map = [('CM', 900), ('CD', 400), ('XC', 90), ('XL', 40), ('IX', 9), ('IV', 4),
                ('M', 1000), ('D', 500), ('C', 100), ('L', 50), ('X', 10), ('V', 5), ('I', 1)]
    s_left = s.upper()  # remaining string...
    result = 0
    while len(s_left) > 0:
        for r, n in num_map:
            if s_left.startswith(r):
                result = result + n
                s_left = s_left[len(r):]
                break
        else:  # did not break
            raise ValueError(f'from_roman_numeral(s) with invalid roman numeral string: s={s!r}')
    return result



"""
File Purpose: logging, recording info about state of code, etc.

For reproducibility in the future.
E.g., with results, also save git hash for current version of the code.
"""

import datetime
import os
import subprocess

from .sentinels import UNSET


''' --------------------- git_hash --------------------- '''

def git_hash_local(*, default=UNSET):
    '''returns the git commit hash for current git HEAD within the local directory

    default: any object
        if not UNSET, and can't get hash (but filepath does exist),
        return default & print message, instead of crashing if can't get hash.
    '''
    try:
        return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()
    except subprocess.CalledProcessError:
        if default is not UNSET:
            print(f'(failed to get git hash; returning default. At dir={repr(os.getcwd())})')
            return default
        raise

def git_hash(module_or_path='.', *, default=UNSET):
    '''returns the git commit hash for current git HEAD at the provided module or path.
    if string, treat it as a path; if module object, use path to module.__file__.
    module_or_path: string or module object, default '.'
        place to get git hash from.
        string --> path=module_or_path. If not a directory, also try os.path.dirname(path)
        module --> module=module_or_path. path=os.path.dirname(module.__file__)
    default: any object
        if provided, and can't get hash, return default & print message.
    '''
    try:
        if isinstance(module_or_path, str):
            path = module_or_path
        else:
            module = module_or_path
            path = module.__file__
        if not os.path.isdir(path):
            path = os.path.abspath(path)
            dirpath = os.path.dirname(path)
            if not os.path.isdir(dirpath):
                raise FileNotFoundError(f'{repr(path)} (not an existing directory). (\nalso {repr(dirpath)})')
            path = dirpath
        try:
            cwd0 = os.getcwd()
            os.chdir(path)
            return git_hash_local(default=default)
        finally:
            os.chdir(cwd0)
    except Exception:
        if default is not UNSET:
            print(f'(failed to get git hash; returning default. For input={repr(module_or_path)})')
            return default
        raise

def git_hash_here(globals_):
    '''returns git commit hash for the __file__ in the namespace where this function is called.'''
    return git_hash(globals_['__file__'])

def git_hash_PlasmaCalcs():
    '''returns the git commit hash for current git HEAD in PlasmaCalcs'''
    return git_hash_here(globals())


''' --------------------- PlasmaCalcs version --------------------- '''

def _PlasmaCalcs_version():
    '''returns the current version of PlasmaCalcs, if available. Else, return None.'''
    try:
        from .. import __version__
    except ImportError:
        return None
    else:
        return __version__


''' --------------------- datetime --------------------- '''

def datetime_now():
    '''current datetime as string according to user's computer. Include timezone.
    Uses "isoformat", with "_" separator: YYYY-MM-DD_HH:MM:SS.ssssss+HH:MM
    Original datetime object can be recovered via `datetime.datetime.fromisoformat(str)`
    '''
    return datetime.datetime.now().astimezone().isoformat(sep='_')


''' --------------------- PlasmaCalcs version --------------------- '''

def code_snapshot_info():
    '''returns dict including info about right now, and the version of the code being used.

    Includes as many of the following as possible (uses None if info unavailable due to crash):
        'pc__version__': PlasmaCalcs version
        'pc__commit_hash': git hash for PlasmaCalcs
        'datetime': current datetime, in isoformat
    '''
    result = {}
    try:
        pc_version = _PlasmaCalcs_version()
    except Exception:
        pc_version = None
    result['pc__version__'] = pc_version
    try:
        pc_hash = git_hash_PlasmaCalcs()
    except Exception:
        pc_hash = None
    result['pc__commit_hash'] = pc_hash
    try:
        datetime = datetime_now()
    except Exception:
        datetime = None
    result['datetime'] = datetime
    return result



"""
File Purpose: QuantTree - tree of LoadableQuantity objects.
"""
import contextlib

from .quantity_tools import MatchedQuantity, MatchedVar, MatchedPattern
from ..errors import InputError
from ..tools import Tree


''' --------------------- QuantTree --------------------- '''

class QuantTree(Tree):
    '''Tree of MatchedQuantity objects.
    See QuantTree.display() for display options.
    Most common use-case:
        QuantTree.from_quantity_loader(quantity_loader, var)
    will tell the QuantTree associated with loading that var from quantity_loader.
    Result might depend on current values of quantity_loader attributes.
    '''
    _assert_type = MatchedQuantity  # assert isinstance(obj, _assert_type) when creating new nodes.

    @classmethod
    def from_quantity_loader(cls, quantity_loader, var, *, missing_ok=False, **kw_call_options):
        '''returns a QuantTree from a QuantityLoader and var name.
        This is the tree of MatchedQuantity objects which will all be loaded if loading var.
            (MatchedQuantity objects consist of a var, a LoadableQuantity, and possibly a _match.)

        quantity_loader: QuantityLoader instance or subclass
            the loader containing info about var dependencies.
            if class, might not be possible to get as many vars' dependencies
                (and will crash if need to know current values of attributes, to determine dependency list)
        missing_ok: bool
            whether to be lenient sometimes when missing details that would allow to fully determine deps.
            see help(MatchedQuantity.dep_vars) for more details.

        additional kwargs appearing in quantity_loader.kw_call_options()
            will be passed to quantity_loader.using(**kw_call_options), if quantity_loader is a QuantityLoader instance.
            If any additional kwargs don't appear in kw_call_options(), crash.
        '''
        if isinstance(quantity_loader, type):
            using = None
        else:
            using = quantity_loader._pop_kw_call_options(kw_call_options)
            if kw_call_options:
                errmsg = (f'unexpected kwargs: {kw_call_options!r}.\n'
                          f'See quantity_loader.kw_call_options() for allowed options.')
                raise InputError(errmsg)
        with contextlib.ExitStack() as context_stack:
            if using:
                context_stack.enter_context(quantity_loader.using(**using))
            matched_quant = quantity_loader.match_var(var)
            tree = cls(matched_quant)  # <-- root node of tree
            dep_vars = matched_quant.dep_vars(quantity_loader=quantity_loader, missing_ok=missing_ok)
            for dep_var in dep_vars:
                if isinstance(dep_var, str):
                    dep_tree = cls.from_quantity_loader(quantity_loader, dep_var)
                elif isinstance(dep_var, tuple):
                    var, using = dep_var
                    with quantity_loader.using(**using):
                        dep_tree = cls.from_quantity_loader(quantity_loader, var)
                else:
                    raise TypeError(f'expected dep_var to be str or tuple, but got {dep_var!r}.')
                tree.add_child(dep_tree)
        return tree
        
    def loading_dims(self):
        '''return a list of all the dims for loading across, as implied by this tree.

        This is the set of quantity.dims for all quantities in this tree (self and all descendants), but:
            - if quantity.ignores_dims is non-empty, exclude those dims for that node and all its descendants.
            - if quantity.reduces_dims is non-empty, exclude those dims for the result
                (even if a sybling node has a quantity that uses those dims).
        '''
        # maintain dims order by using a list instead of a set. (Not worried about speed here.)
        # dims from here
        dims = self.obj.dims
        if dims is None:
            result = []
        else:
            result = list(dims)  # copy of dims, as a list.
        # dims from children
        for child in self.children:
            child_dims = child.loading_dims()
            for dim in child_dims:
                if dim not in result:
                    result.append(dim)
        # remove dims from any ignores_dims here.
        #  (Descendants' ignores_dims were already excluded by their loading_dims())
        for dim in self.obj.ignores_dims:
            if dim in result:
                result.remove(dim)
        # remove dims from any reduces_dims anywhere in the whole tree.
        for node in self.flat(include_self=True):
            for dim in node.obj.reduces_dims:
                if dim in result:
                    result.remove(dim)
        # return result
        return result

    def result_dims(self):
        '''return a list of all the dims the result would have, as implied by this tree.

        This is the set of quantity.dims for all quantities in this tree (self and all descendants), but:
            - if quantity.ignores_dims is non-empty, exclude those dims for that node and all its descendants.
            - if quantity.reduces_dims is non-empty, exclude those dims for that node and all its descendants,
                (but not from the whole tree. See self.loading_dims() to exclude those from the whole tree.)
        '''
        # maintain dims order by using a list instead of a set. (Not worried about speed here.)
        # dims from here
        dims = self.obj.dims
        if dims is None:
            result = []
        else:
            result = list(dims)  # copy of dims, as a list.
        # dims from children
        for child in self.children:
            child_dims = child.result_dims()
            for dim in child_dims:
                if dim not in result:
                    result.append(dim)
        # remove dims from any ignores_dims or reduces_dims here.
        #  (Descendants' ignores_dims and reduces_dims were already excluded by their result_dims().)
        for dim in [*self.obj.ignores_dims, *self.obj.reduces_dims]:
            if dim in result:
                result.remove(dim)
        # return result
        return result

    def flat_branches_until_vars(self, *, include_self=False):
        '''iterator over self and all descendants, but no descendants of MatchedVar nodes.'''
        return self.flat_branches_until(lambda node: isinstance(node.obj, MatchedVar), include_self=include_self)



"""
File Purpose: errors in PlasmaCalcs
"""


''' --------------------- Misc Errors --------------------- '''

class BindingError(ValueError):
    '''error indicating an issue with binding; see tools.binding.'''
    pass

class ImportFailedError(ImportError):
    '''error indicating that an import failed in the past, which is why a module cannot be accessed now.'''
    pass

class TimeoutError(Exception):
    '''error indicating a timeout.'''
    pass

class FileAmbiguityError(OSError):
    '''error indicating some ambiguity due to which files exist or don't exist.'''
    pass


''' --------------------- Quantity Calculation Errors --------------------- '''

class QuantCalcError(ValueError):
    '''error in quantity calculation.'''
    pass

class FormulaMissingError(QuantCalcError, NotImplementedError):
    '''error in locating formula to calculate quant.'''
    pass

class OverrideNotApplicableError(QuantCalcError):
    '''error indicating that an override function is not applicable.'''
    pass

class CacheNotApplicableError(QuantCalcError):
    '''error indicating there is no relevant value in the cache.'''
    pass

class UnitsError(QuantCalcError):
    '''error related to units.'''
    pass

class UnitsUnknownError(UnitsError):
    '''error related to units being unknown.'''
    pass

class FileContentsError(QuantCalcError):
    '''error related to contents of a file being incorrect or unexpected.'''
    pass

class ValueLoadingError(FileContentsError):
    '''error related to loading a value from a file, but NOT just because it's not yet implemented.
    E.g., when it is known that value "should" exist, but it turns out to not exist in the file.
    '''
    pass

class FileContentsMissingError(FileContentsError):
    '''error indicating missing information in a file.'''
    pass

class FileContentsConflictError(FileContentsError):
    '''error related to conflicting information in a file.'''
    pass

class LoadingNotImplementedError(FormulaMissingError):
    '''error indicating failed to implement how to load these values from files.'''
    pass

class SetvarNotImplementedError(FormulaMissingError):
    '''error indicating failed to implement setvar, e.g. for a particular var.'''
    pass

class TypevarNanError(FormulaMissingError):
    '''error indicating a typevar is nan,
    probably due to conditions making it impossible to load the corresponding var.
    '''
    pass

class MemorySizeError(QuantCalcError):
    '''error indicating size of data is too large.'''
    pass

class CollisionsModeError(QuantCalcError):
    '''error indicating an issue with the mode of collisions.'''
    pass


''' --------------------- InputErrors --------------------- '''

class InputError(TypeError):
    '''error indicating something is wrong with the inputs, e.g. to a function.'''
    pass

class InputConflictError(InputError):
    '''error indicating two or more inputs provide conflicting information.
    E.g. foo(lims=None, vmin=None, vmax=None) receiving lims=(1,7), vmin=3, might raise this error,
    if the intention is for vmin and vmax to be aliases to lims[0] and lims[1].
    '''
    pass

class InputMissingError(InputError):
    '''error indicating that an input is missing AND doesn't have an appropriate default value.
    E.g. default=None; def foo(kwarg=None): if kwarg is None: kwarg=default; but foo expects non-None value.
    '''
    pass

class QuantInfoMissingError(InputError, QuantCalcError):
    '''error indicating that info is missing during a quantity calculation.
    E.g., cross table missing when attempting to read cross section.
    '''
    pass


''' --------------------- Dimensions --------------------- '''

class DimensionError(Exception):
    '''error indicating some issue with a dimension'''
    pass

class DimensionalityError(DimensionError):
    '''error indicating dimensionality issue, e.g. wrong number of dimensions'''
    pass

class DimensionSizeError(DimensionError):
    '''error indicating a dimension is the wrong size.'''
    pass

class DimensionKeyError(KeyError, DimensionError):
    '''error indicating missing value of a dimension (e.g. FluidKeyError, SnapKeyError).'''
    def __str__(self):
        '''use standard error string. Avoid KeyError string which uses repr(message).'''
        return super(KeyError, self).__str__()

class DimensionValueError(ValueError, DimensionError):
    '''error indicating some incompatibility regarding the value of a dimension (e.g. SnapValueError)'''
    pass

class DimensionAttributeError(AttributeError, DimensionError):
    '''error indicating some issue with an attribute of a dimension, probably that it doesn't exist.'''
    pass

class ComponentKeyError(DimensionKeyError):
    '''error indicating component key not found.'''
    pass

class ComponentValueError(DimensionValueError):
    '''error indicating some issue with value for a Component.'''
    pass

class FluidKeyError(DimensionKeyError):
    '''error indicating Fluid key not found.'''
    pass

class FluidValueError(DimensionValueError):
    '''error indicating some issue with value for a Fluid.'''
    pass

class SnapKeyError(DimensionKeyError):
    '''error indicating Snap key not found.'''
    pass

class SnapValueError(DimensionValueError):
    '''error indicating some issue with value for a Snap.'''
    pass

class RunKeyError(DimensionKeyError):
    '''error indicating Run key not found.'''
    pass

class RunValueError(DimensionValueError):
    '''error indicating some issue with value for a Run.'''
    pass


''' --------------------- PlottingErrors --------------------- '''

class PlottingError(ValueError):
    '''error indicating an issue with plotting'''
    pass

class MappableNotFoundError(PlottingError):
    '''error indicating mappable not found. E.g., might be raised during colorbar()'''
    pass

class PlottingAmbiguityError(PlottingError):
    '''error indicating an ambiguity with plotting, e.g. multiple images found when looking for only 1.'''
    pass

class PlottingNotImplementedError(PlottingError, NotImplementedError):
    '''error indicating some aspect of plotting is not implemented'''
    pass

class TooManyPlottablesError(PlottingError):
    '''error indicating too many things to make a plot.'''
    pass

class TooManySubplotsError(TooManyPlottablesError):
    '''error indicating too many subplots.'''
    pass

class PlotSettingsError(PlottingError):
    '''error indicating an issue with plot settings.'''
    pass

class PlottingNframesUnknownError(PlottingNotImplementedError):
    '''error indicating number of frames unknown for plotting a movie.'''
    pass


''' --------------------- Serializing --------------------- '''

class SerializationError(ValueError):
    '''error indicating an issue related to serialization or deserialization.'''
    pass

class SerializingError(SerializationError):
    '''error indicating an issue when serializing.'''
    pass

class DeserializingError(SerializationError):
    '''error indicating an issue when deserializing.'''
    pass


''' --------------------- Subsampling --------------------- '''

class SubsamplingError(ValueError):
    '''error indicating an issue with subsampling.'''
    pass

class SubsamplingFormatError(SubsamplingError):
    '''error indicating an issue with the format of subsampling.'''
    pass

class SubsamplingNotFoundError(SubsamplingError):
    '''error indicating that something related to subsampling was expected but not found.'''
    pass



"""
Package Purpose: Miscellaneous quality-of-life routines for Object Oriented Programming tasks.
"""

from .binding import (
    bind_to, Binding,
)
from .manage_attrs import (
    maintaining_attrs, using_attrs,
    maintain_attrs, use_attrs,
    MaintainingAttrs, UsingAttrs,
    UsingAttrsSignaled,
    IncrementableAttrManager,
)
from .oop_misc import (
    apply,
    MetaClsRepr,
)


"""
File Purpose: tools related to xarray coords
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_dimensions import xarray_promote_dim, xarray_ensure_dims
from .xarray_misc import xarray_dtype_object_to_str
from ..arrays import ndindex_array
from ..math import float_rounding as math_float_rounding
from ...errors import (
    DimensionalityError, DimensionValueError, DimensionKeyError,
    InputError, InputConflictError, InputMissingError,
)


# note: see xarray_indexing.py for xarray_cmin, xarray_cmax,
#   xarray_min_coord_where, and xarray_max_coord_where.

''' --------------------- Coords --------------------- '''

@pcAccessor.register('nondim_coords')
def xarray_nondim_coords(array, *, scalars_only=False, item=False, _sort=True):
    '''returns dict of {coord name: coord.values} for all non-dimension coords (not in array.dims).
    scalars_only: bool
        whether to only include coords with ndim==0.
    item: bool
        whether to use coord.item() for scalars.

    _sort: bool
        whether to sort, if any '{coord}_index' coords appear consecutively,
        (when doing the default array.coords.items() order), sort those alphabetically.
        If their non-index counterparts also appear consecutively, sort those too.
        E.g. order: 'C', 'A', 'B', 'other_index', 'D', 'C_index', 'A_index', 'D_index', 'B_index', 'E',
           becomes: 'A', 'B', 'C', 'other_index', 'D', 'A_index', 'B_index', 'C_index', 'D_index', 'E'.
    '''
    result = {cname: coord.values for cname, coord in array.coords.items() if cname not in array.dims}
    if scalars_only:
        result = {cname: val for cname, val in result.items() if np.ndim(val) == 0}
    if item:
        result = {cname: val.item() if val.ndim==0 else val for cname, val in result.items()}
    if not _sort:
        return result
    else:  # _sort=True
        # sorting (stylistic choice only. Might affect plot title order.)
        # [TODO] make an xarray_sort_coords function to produce a standard order, instead of writing it here.
        order = list(result.keys())
        new_order = []
        grouped_indexes = []
        iter_order = iter(order)
        for cname in iter_order:
            group = []
            cname2 = None
            if cname.endswith('_index'):
                group.append(cname)
                for cname2 in iter_order:
                    if cname2.endswith('_index'):
                        group.append(cname2)
                    else:
                        break
            if len(group) > 1:
                group = sorted(group)
                grouped_indexes.append(group)
                new_order.extend(group)
                new_order.append(cname2)
            elif cname2 is not None:  # found an index coord, but only 1 consecutively.
                new_order.append(cname)
                new_order.append(cname2)
            else:  # not an index coord.
                new_order.append(cname)
            order = new_order
        # (look for non-index counterparts to sort)
        if len(grouped_indexes) > 0:
            nonindex_groups = [[cname[:-len('_index')] for cname in group] for group in grouped_indexes]
            new_order = []
            iter_order = iter(order)
            for cname in iter_order:
                group = []
                for ng in nonindex_groups:
                    if cname in ng:
                        group.append(cname)
                        break
                else:  # this cname doesn't have an index group.
                    new_order.append(cname)
                    continue
                for cname2 in iter_order:
                    if cname2 in ng:
                        group.append(cname2)
                    else:
                        new_order.extend(sorted(group))
                        new_order.append(cname2)
                        break
            order = new_order
        return {cname: result[cname] for cname in order}

@pcAccessor.register('dims_coords')
def xarray_dims_coords(array, *, include_dims_as_coords=True):
    '''returns dict of {dim name: [coord name for all coords with this dim]}.
    result[()] will be list of all scalar coords (ndim=0 so no associated dims).
    coords associated with multiple dims will appear in multiple places in the result.

    include_dims_as_coords: bool
        whether to include dims as coord names in the result.
        Dims with no same-named coord will appear in appropriate place in result.
    '''
    result = dict()
    unused_dims = set(array.dims)
    for cname, coord in array.coords.items():
        unused_dims -= set([cname])
        if len(coord.dims) == 0:
            result.setdefault((), []).append(cname)
        for dim in coord.dims:
            result.setdefault(dim, []).append(cname)
    if include_dims_as_coords and unused_dims:
        for dim in unused_dims:
            result.setdefault(dim, []).append(dim)
    return result

@pcAccessor.register('assign_self_as_coord', totype='array')
def xarray_assign_self_as_coord(array):
    '''return copy of array with coord named array.name, equal to array values.
    Equivalent: array.assign_coords({array.name: array})
    '''
    if array.name is None:
        raise InputError('assign_self_coord expects non-None array.name')
    return array.assign_coords({array.name: array})

@pcAccessor.register('fill_coords')
def xarray_fill_coords(array, dim=None, *, need=None):
    '''return copy of array with coords filled for indicated dims.
    (if all indicated dims have coords already, return original array, not a copy.)
    E.g. array with dim_1 length 50 but no coords
        --> result is just like array but has dim_1 coords = np.arange(50)

    dim: None, str, or iterable of strs
        dims for which to consider filling coords. None --> array.dims.
    need: None, str, or iterable of str or None
        coords which the result must contain.
        if any of these look like '{coord}_index', 'log_{coord}', or 'str_{coord}',
            create them via xarray_index_coords, xarray_log_coords, or xarray_str_coords.
    '''
    if dim is None: dim = array.dims
    elif isinstance(dim, str): dim = [dim]
    to_assign = {}
    for d in dim:
        if d not in array.coords:
            to_assign[d] = array[d]  # array[d] == np.arange(len(array[d]))
    if to_assign:
        array = array.assign_coords(to_assign)
    if need is not None:
        if isinstance(need, str): need = [need]
        for cname in need:
            if cname is None or cname in array.coords:
                continue
            elif cname.endswith('_index'):
                array = xarray_index_coords(array, coords=cname[:-len('_index')], drop=False)
            elif cname.startswith('log_'):
                array = xarray_log_coords(array, coords=cname[len('log_'):], drop=False)
            elif cname.startswith('str_'):
                array = xarray_str_coords(array, coords=cname[len('str_'):], drop=False)
            else:
                errmsg = (f'cname={cname!r} in need={need} but not in '
                          f'array.dims ({list(array.dims)}) nor array.coords ({list(array.coords)}).\n'
                          '(Also no implied coord to create; cname not like "{coord}_index" or "log_{coord}".')
                raise InputConflictError(errmsg)
    return array

@pcAccessor.register('index_coords')
def xarray_index_coords(array, coords=None, newname='{coord}_index', *,
                        drop=False, promote=False, exist_ok=False, max_ndim=None):
    '''return copy of array with coord_index coords telling np.arange() for each coord.
    0D coords' index will be 0 if provided explicitly in coords input, else ignored.
    1D coords' index will be np.arange, e.g. coord_index[i] == i.
    2D+ coords' index will be reshaped np.ndindex, such that coord_index[i,j] == (i,j).

    coords: None or iterable of strs
        if None, use all coords and dims which don't already have coord_index.
        (e.g. 'fluid' --> make 'fluid_index', unless 'fluid_index' already exists.)
    newname: str
        string for new (index) coord names: newname.format(coord=coord).
        Default: '{coord}_index'. To keep original names, use '{coord}'
    drop: bool
        whether to drop original coords after creating coord_index coords.
        (e.g. 'fluid' --> drop 'fluid' after making 'fluid_index')
    promote: bool
        whether to promote all non-dim new index coords to dimensions.
        if True, xarray_promote_dim for all new index coords.
    exist_ok: bool
        whether it is okay if newname for coord (e.g. 'fluid_index') already exists.
        True --> replace existing coord with new coord_index.
    max_ndim: None or int
        if not None, skip any coords with ndim > max_ndim.
        E.g. max_ndim=1 prevents making indexes for coords with ndim>=2.
    '''
    if coords is None:
        nonscalar_coords = [cname for cname, cval in array.coords.items() if cval.ndim >= 1]
        coords = list(set(nonscalar_coords).union(array.dims))
    elif isinstance(coords, str):
        coords = [coords]
    newnames = {cname: newname.format(coord=cname) for cname in coords}
    if len(set(newnames.values())) != len(newnames):  # crash if any newname apears multiple times:
        raise InputConflictError(f'old to new name map must have unique values but got: {newnames!r}')
    if not exist_ok:
        existing = set(array.coords).intersection(newnames.values())
        if existing:
            raise InputConflictError(f'some new coord names already exist in array.coords: {existing!r}')
    # computing new coords
    to_assign = {}
    for cname in coords:
        cc = array.coords[cname]
        if (max_ndim is not None) and cc.ndim > max_ndim:
            continue
        if cc.ndim == 0:
            index = np.array(0, dtype=np.min_scalar_type(0))
        else:
            index = ndindex_array(cc.shape)
        newval = cc.copy(data=index)
        if drop:
            newval = newval.drop_vars(coords, errors='ignore')  # 'ignore' if newval missing any coords
        to_assign[newnames[cname]] = newval
    # creating result
    if drop:
        array = array.drop_vars(coords, errors='ignore')
    result = array.assign_coords(to_assign)
    if promote:
        result = xarray_ensure_dims(result, list(newnames.values()))
    return result

@pcAccessor.register('promote_index_coords')
def xarray_promote_index_coords(array, coords=None):
    '''promote to dims all '{cname}_index' coords for which '{cname}' is associated with a 1D dim.
    
    coords: None, str, or list of str
        coords (with or without '_index' suffix) to consider promoting to dims.
        None --> all index coords.
        str --> treat as list of 1 element.
    '''
    if coords is None:
        coords = [c for c in array.coords if c.endswith('_index')]
    elif isinstance(coords, str):
        coords = [coords]
    index_names = [c if c.endswith('_index') else f'{c}_index' for c in coords]
    orig_names = [c[:-len('_index')] for c in index_names]
    for oo, ii in zip(orig_names, index_names):
        if ii in array.coords:
            cval = array.coords[ii]
            if cval.ndim == 1:
                array = xarray_promote_dim(array, ii)
    return array

@pcAccessor.register('demote_index_coords')
def xarray_demote_index_coords(array, coords=None):
    '''demote '{cname}_index' dims to coords by promoting '{cname}' coords to dim instead.
    Skips index dims for which associated '{cname}' coord does not exist.

    coords: None, str, or list of str
        coords (with or without '_index' suffix) to consider demoting to dims.
        None --> all index coords.
        str --> treat as list of 1 element.
    '''
    if coords is None:
        coords = [c for c in array.coords if c.endswith('_index')]
    elif isinstance(coords, str):
        coords = [coords]
    index_names = [c if c.endswith('_index') else f'{c}_index' for c in coords]
    orig_names = [c[:-len('_index')] for c in index_names]
    for oo, ii in zip(orig_names, index_names):
        if oo in array.coords:
            cval = array.coords[oo]
            if cval.ndim == 1:
                array = xarray_promote_dim(array, oo)
    return array

@pcAccessor.register('scale_coords')
def xarray_scale_coords(array, scale=None, *, missing_ok=True, **scale_as_kw):
    '''return copy of array with coords multiplied by scale.
    scale: None or dict of {coord: scale}
        dict --> multiply each coord by the corresponding number.
        None --> provide as kwargs (scale_as_kw) instead.
    scale_as_kw: if scale is None, can provide scale dict as kwargs instead.
    missing_ok: bool
        whether it is okay if some coords are missing (if yes, skip missing coords).
    '''
    if scale is None and len(scale_as_kw) == 0:
        raise InputMissingError('must provide either "scale" or "scale_as_kw".')
    if scale is not None and len(scale_as_kw) > 0:
        raise InputConflictError('cannot provide both "scale" and "scale_as_kw".')
    if scale is None:
        scale = scale_as_kw
    assign_coords = {}
    for cname, cscale in scale.items():
        try:
            cvals = array.coords[cname]
        except KeyError:
            if not missing_ok:
                raise DimensionKeyError(f'coord={cname!r} not found in array.coords.') from None
            continue
        assign_coords[cname] = cvals * cscale
    return array.assign_coords(assign_coords)

@pcAccessor.register('shift_coords')
def xarray_shift_coords(array, shift=None, *, missing_ok=True, **shift_as_kw):
    '''return copy of array with coords shifted by shift.
    shift: None or dict of {coord: shift}
        dict --> shift each coord by the corresponding number.
        None --> provide as kwargs (shift_as_kw) instead.
    shift_as_kw: if shift is None, can provide shift dict as kwargs instead.
    missing_ok: bool
        whether it is okay if some coords are missing (if yes, skip missing coords).
    '''
    if shift is None and len(shift_as_kw) == 0:
        raise InputMissingError('must provide either "shift" or "shift_as_kw".')
    if shift is not None and len(shift_as_kw) > 0:
        raise InputConflictError('cannot provide both "shift" and "shift_as_kw".')
    if shift is None:
        shift = shift_as_kw
    assign_coords = {}
    for cname, cshift in shift.items():
        try:
            cvals = array.coords[cname]
        except KeyError:
            if not missing_ok:
                raise DimensionKeyError(f'coord={cname!r} not found in array.coords.') from None
            continue
        assign_coords[cname] = cvals + cshift
    return array.assign_coords(assign_coords)

@pcAccessor.register('mod_coords')
def xarray_mod_coords(array, mod=None, *, missing_ok=True, modshift=0, **mod_as_kw):
    '''return copy of array with coords modded (via np.mod) by `mod`.
    mod: None or dict of {coord: mod}
        dict --> mod each coord by the corresponding number.
        None --> provide as kwargs (mod_as_kw) instead.
    mod_as_kw: if mod is None, can provide mod dict as kwargs instead.
    modshift: value
        shift each coord by modshift before modding, then -modshift after modding.
        E.g. if modshift=90, mod=180, then modded coord = np.mod(coord+90, 180)-90.
    missing_ok: bool
        whether it is okay if some coords are missing (if yes, skip missing coords).
    '''
    if mod is None and len(mod_as_kw) == 0:
        raise InputMissingError('must provide either "mod" or "mod_as_kw".')
    if mod is not None and len(mod_as_kw) > 0:
        raise InputConflictError('cannot provide both "mod" and "mod_as_kw".')
    if mod is None:
        mod = mod_as_kw
    assign_coords = {}
    for cname, cmod in mod.items():
        try:
            cvals = array.coords[cname]
        except KeyError:
            if not missing_ok:
                raise DimensionKeyError(f'coord={cname!r} not found in array.coords.') from None
            continue
        assign_coords[cname] = np.mod(cvals + modshift, cmod) - modshift
    return array.assign_coords(assign_coords)

@pcAccessor.register('log_coords')
def xarray_log_coords(array, coords=None, newname='log_{coord}', *,
                      base=10, drop=True, promote=False):
    '''return copy of array with coords replaced by log coords (& renamed to log_coord)

    coords: None, str, or iterable of strs
        coords to replace with log. None --> all coords.
    newcoord: str
        string for new (logged) coord names: newcoord.format(coord=coord).
        Default: 'log_{coord}'. To keep original names, use '{coord}'
    base: number or 'e'
        log in this base. default 10.
        if not 10, result.assign_attrs({'log_base': base}).
    drop: bool
        whether to drop original coords' values.
        True --> drop original coords.
        False --> add new coords but do not adjust original coords.
    promote: bool
        whether to promote all non-dim new log coords to dimensions.
        if True, xarray_promote_dim for all new log coords.
    '''
    # bookkeeping
    if coords is None:
        coords = array.coords
    elif isinstance(coords, str):
        coords = [coords]
    newnames = {cname: newname.format(coord=cname) for cname in coords}
    if len(set(newnames.values())) != len(newnames):  # crash if any newname apears multiple times:
        raise InputConflictError(f'old to new name map must have unique values but got: {newnames!r}')
    if base == 10:
        f_log = np.log10
    else:
        array = array.assign_attrs({'log_base': base})
        if base == 'e':
            f_log = np.log
        elif base == 2:
            f_log = np.log2
        else:
            f_log = lambda x: np.log(x) / np.log(base)
    # computing new coords
    newcoords = {}
    for cname in coords:
        cvals = array.coords[cname]
        if drop:
            cvals = cvals.drop_vars(coords, errors='ignore')
        newvals = f_log(cvals)
        newcoords[newnames[cname]] = newvals
    # creating result appropriately
    if drop:
        array = array.drop_vars(coords, errors='ignore')
    result = array.assign_coords(newcoords)
    if promote:
        result = xarray_ensure_dims(result, list(newnames.values()))
    return result

@pcAccessor.register('str_coords')
def xarray_str_coords(array, coords=None, newname='str_{coord}', *, drop=True, promote=False,
                      missing_ok=False):
    '''return copy of array with coords replaced by str coords (& renamed to str_coord)
    E.g. array.pc.str_coords('fluid') --> result['str_fluid'] == str(f) for f in array['fluid']

    coords: None, str, or iterable of strs
        coords to replace with str. None --> all coords.
    newcoord: str
        string for new (converted-to-str) coord names: newcoord.format(coord=coord).
        Default: 'str_{coord}'. To keep original names, use '{coord}'
    drop: bool
        whether to drop original coords' values.
        True --> drop original coords.
        False --> add new coords but do not adjust original coords.
    promote: bool
        whether to promote all non-dim new str coords to dimensions.
        if True, xarray_promote_dim for all new str coords.
    missing_ok: bool
        whether it is okay for some provided `coords` to be missing from array.coords.
    '''
    # bookkeeping
    if coords is None:
        coords = array.coords
    elif isinstance(coords, str):
        coords = [coords]
    if missing_ok:
        coords = [c for c in coords if c in array.coords]  # skips missing coords.
    newnames = {cname: newname.format(coord=cname) for cname in coords}
    if len(set(newnames.values())) != len(newnames):
        raise InputConflictError(f'old to new name map must have unique values but got: {newnames!r}')
    # computing
    newcoords = {}
    for cname in coords:
        cvals = array.coords[cname]
        if drop:
            cvals = cvals.drop_vars(coords, errors='ignore')
        if cvals.dtype == object:
            newvals = xarray_dtype_object_to_str(cvals)
        else:
            newvals = cvals.astype(str)
        newcoords[newnames[cname]] = newvals
    # creating result
    if drop:
        array = array.drop_vars(coords, errors='ignore')
    result = array.assign_coords(newcoords)
    if promote:
        result = xarray_ensure_dims(result, list(newnames.values()))
    return result

@pcAccessor.register('is_sorted', totype='array')
def xarray_is_sorted(array, *, increasing=True):
    '''returns whether array is sorted; array must be 1D.

    increasing: bool
        True --> check for increasing order. vals[i] <= vals[i+1]
        False --> check for decreasing order. vals[i] >= vals [i+1]
    '''
    if array.ndim != 1:
        raise DimensionalityError('is_sorted expects 1D array.')
    vals = array.data
    if increasing:
        return np.all(vals[:-1] <= vals[1:])
    else:
        return np.all(vals[:-1] >= vals[1:])


''' --------------------- Coord Math --------------------- '''

@pcAccessor.register('get_dx_along')
def xarray_get_dx_along(array, coord, *, atol=0, rtol=1e-5, float_rounding=False):
    '''returns number equal to the diff along array.coords[coord], after checking that it is constant.
    result will be a single number, equal to array.coords[coord].diff(coord)[0].item().

    (Technically, also promotes coord to dim during calculations if coord was a non-dimension coordinate.)
    
    before returning result, ensure that np.allclose(array.diff(dim), atol=atol, rtol=rtol);
        if that fails, raise DimensionValueError.

    float_rounding: bool
        if True, re-create floating point result if it seems to be wrong by only a small amount,
        e.g. 0.20000000001 --> float(0.2); 0.39999999999 --> float(0.4); 0.123456781234 --> unchanged
        This sometimes improves "exact" float comparisons, if float was input from a string.
        See tools.float_rounding for more details.
    '''
    carr = array.coords[coord]
    carr = xarray_promote_dim(carr, coord)
    diff = carr.diff(coord)
    if len(diff) == 0:
        raise DimensionValueError(f'expected non-empty diff({coord!r})')
    result = diff[0].item()
    if not np.allclose(diff, result, atol=atol, rtol=rtol):
        errmsg = f'expected evenly-spaced coordinates along coord {coord!r}, but got diff={diff}'
        raise DimensionValueError(errmsg)
    if float_rounding:
        result = math_float_rounding(result)
    return result

@pcAccessor.register('differentiate')
def xarray_differentiate(array, coord, *, keep_attrs=True, **kw__differentiate):
    '''differentiate array along coord, treating array like it is an xarray.DataArray.
    more lenient than xarray.DataArray.differentiate;
        returns 0 if can't differentiate along coord (due to coord having size 1 or not existing.)

    keep_attrs: bool
        whether to copy attrs from array into the result. Default True.

    requires that array.coords and array.differentiate exist, otherwise raises AttributeError.
    '''
    coords = array.coords
    try:
        coords_x = coords[coord]
    except KeyError:
        return xr.zeros_like(array)
    size_x = np.size(coords_x)
    if size_x <= 1:
        return xr.zeros_like(array)
    else:
        result = array.differentiate(coord, **kw__differentiate)
        if keep_attrs:
            result = result.assign_attrs(array.attrs.copy())
        return result



"""
File Purpose: calculations involving values from multiple runs,
or multiple analogous groupings of non-simulation data.
"""

from .multi_calculator import MultiCalculator
from .multi_run_tools import canon_runs



"""
File Purpose: loading quantities related to addons but specific to eppic.
"""

from ...addons import ADDON_LOADERS
from ...quantities import QuantityLoader

loaded_tfbi = any(_cls.__name__=='TfbiLoader' for _cls in ADDON_LOADERS)

class EppicAddonQuantityLoader(QuantityLoader):
    '''Loader for quantities related to addons but specific to eppic.'''
    if loaded_tfbi:
        @known_var(deps=['tfbi_vs_EBspeed'])
        def get_inputs_tfbi_vs_EBspeed(self):
            '''get tfbi_vs_EBspeed at self.set_vars_from_inputs().
            Makes a copy of self to do this computation, to avoid altering original self.
            [TODO] update this if internal methods for this don't alter the copy anymore
            '''
            ec = self.copy()
            ec.set_vars_from_inputs()
            return ec('tfbi_vs_EBspeed')



"""
File Purpose: xarray math with error propagation.
"""

import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_agg_stats import (
    _paramdocs_aggregator, _xarray_aggregator_setup,
    xarray_stats,
)
from ..docs_tools import format_docstring
from ...errors import InputError, InputConflictError


''' --------------------- mean with error --------------------- '''

@pcAccessor.register('werrmean', totype='array')
@format_docstring(**_paramdocs_aggregator)
def xarray_werrmean(array, dim=None, *, keep=None,
                    promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''returns Dataset of 'mean' and 'std' for array.
    Equivalent: xarray_stats(array, stats=['mean', 'std'], ...)

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.mean and std.
    Can provide skipna=True/False to skip NaNs or not; default True for float dtypes.

    see also: xarray_werradd, xarray_werrsub, xarray_werrmul, xarray_werrdiv.
    Example:
        ds = arr.pc.werrmean().pc.werradd(arr2.pc.werrmean())
        ds['mean'] == arr.mean() + arr2.mean(),
        ds['std'] == sqrt(arr.std()**2 + arr2.std()**2)
    '''
    kw.update(stats=['mean', 'std'],
              dim=dim, keep=keep,
              promote_dims_if_needed=promote_dims_if_needed,
              missing_dims=missing_dims)
    return xarray_stats(array, **kw)

@pcAccessor.register('werr2pmstd', totype='dataset')
def xarray_werr2pmstd(ds, *, keep_mean=False):
    '''return ds but replacing 'std' data_var with 'mean+std' and 'mean-std'.
    ds: xarray.Dataset
        must have 'mean' and 'std' data vars.
    keep_mean: bool
        whether to keep 'mean' data_var in the result.

    see also: xarray_pmstd2werr
    '''
    mean = ds['mean']
    std = ds['std']
    ds = ds.drop_vars('std')
    to_assign = {}
    to_assign['mean+std'] = mean + std
    if keep_mean: to_assign['mean'] = mean
    to_assign['mean-std'] = mean - std
    ds = ds.assign(to_assign)
    return ds

@pcAccessor.register('pmstd2werr', totype='dataset')
def xarray_pmstd2werr(ds):
    '''return ds but replacing 'mean+std' and 'mean-std' data_vars with 'std'.
    ds: xarray.Dataset
        must have 'mean+std', and 'mean-std' data vars.
        if also has 'mean' data var, assert mean = ('mean+std' + 'mean-std') / 2,
            crashing with InputConflictError if not.

    see also: xarray_werr2pmstd
    '''
    mean = (ds['mean+std'] + ds['mean-std']) / 2
    std = ds['mean+std'] - mean
    if 'mean' in ds.data_vars and (not np.allclose(ds['mean'], mean)):
        raise InputConflictError('datavar mean != (mean+std + mean-std) / 2')
    ds = ds.drop_vars(['mean+std', 'mean-std'])
    ds = ds.assign({'std': std})
    return ds

@pcAccessor.register('mean_pm_std', totype='array')
@format_docstring(**_paramdocs_aggregator)
def xarray_mean_pm_std(array, dim=None, *, keep=None,
                       promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''returns Dataset of 'mean+std', 'mean', 'mean-std' for array.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.mean and std.
    Can provide skipna=True/False to skip NaNs or not; default True for float dtypes.
    '''
    array, dim, keep = _xarray_aggregator_setup(array, dim, keep, promote_dims_if_needed, missing_dims)
    mean = array.mean(dim, **kw)
    std = array.std(dim, **kw)
    attrs = array.attrs.copy() if xr.get_options()['keep_attrs'] else None
    results = {'mean+std': mean + std, 'mean': mean, 'mean-std': mean - std}
    return xr.Dataset(results, attrs=attrs)


''' --------------------- math with error propagation --------------------- '''

_paramdocs_werrmath = {
    'A_B': '''xarray.Dataset or DataArray; B can also be any other value.
        Dataset --> mean = ds['mean']; std = ds.get('std', default=0).
        else --> assume it represents 'mean', with std=0.''',
    'require_std': '''bool
        whether to require at least one of A or B to have 'std'.
        if True and std missing from both, crash with InputError.''',
}

def _xarray_werrmath_inputs(A, B, *, require_std=True):
    '''return meanA, stdA, meanB, stdB.
    if 'std' missing from one, assume std=0 for that one.
    if require_std and std missing from both, crash with InputError.
    '''
    found_std = []
    if isinstance(A, xr.Dataset):
        Amean = A['mean']
        Astd = A.get('std', 0)
        found_std.append('A')
    else:
        Amean = A
        Astd = 0
    if isinstance(B, xr.Dataset):
        Bmean = B['mean']
        Bstd = B.get('std', 0)
        found_std.append('B')
    else:
        Bmean = B
        Bstd = 0
    if require_std and not found_std:
        raise InputError('to use werrmath, A and/or B must be a Dataset with "std" data_var.')
    return Amean, Astd, Bmean, Bstd

def _xarray_werrmath_postprocess(A, B, mean, std):
    '''return Dataset of mean and std, copying all other details from A (if Dataset) else B.
    (other details include coords, attrs, and any data_vars unrelated to werrmath.)
    '''
    if isinstance(A, xr.Dataset):
        result = A
    elif isinstance(B, xr.Dataset):
        result = B
    else:
        raise InputError(f'expected A or B to be a Dataset, got {type(A)} and {type(B)}')
    result = result.assign({'mean': mean, 'std': std})
    return result

@pcAccessor.register('werradd')
@format_docstring(**_paramdocs_werrmath)
def xarray_werradd(A, B, *, require_std=True):
    '''returns dataset of 'mean' and 'std' for A + B, including error propagation.
    Assumes independent errors and applies the "standard" formula:
        mean(A + B) = mean(A) + mean(B)
        std(A + B) = sqrt(std(A)**2 + std(B)**2)

    A, B: {A_B}
    require_std: {require_std}
    '''
    Amean, Astd, Bmean, Bstd = _xarray_werrmath_inputs(A, B, require_std=require_std)
    mean = Amean + Bmean
    std = np.sqrt(Astd**2 + Bstd**2)
    return _xarray_werrmath_postprocess(A, B, mean, std)

@pcAccessor.register('werrsub')
@format_docstring(**_paramdocs_werrmath)
def xarray_werrsub(A, B, *, require_std=True):
    '''returns dataset of 'mean' and 'std' for A - B, including error propagation.
    Assumes independent errors and applies the "standard" formula:
        mean(A - B) = mean(A) - mean(B)
        std(A - B) = sqrt(std(A)**2 + std(B)**2)

    A, B: {A_B}
    require_std: {require_std}
    '''
    Amean, Astd, Bmean, Bstd = _xarray_werrmath_inputs(A, B, require_std=require_std)
    mean = Amean - Bmean
    std = np.sqrt(Astd**2 + Bstd**2)
    return _xarray_werrmath_postprocess(A, B, mean, std)

@pcAccessor.register('werrmul')
@format_docstring(**_paramdocs_werrmath)
def xarray_werrmul(A, B, *, require_std=True):
    '''returns dataset of 'mean' and 'std' for A * B, including error propagation.
    Assumes independent errors and applies the "standard" formula:
        z = A * B
        mean(z) = mean(A) * mean(B)
        std(z) = abs(mean(z)) * sqrt((std(A)/mean(A))**2 + (std(B)/mean(B))**2)

    A, B: {A_B}
    require_std: {require_std}
    '''
    Amean, Astd, Bmean, Bstd = _xarray_werrmath_inputs(A, B, require_std=require_std)
    mean = Amean * Bmean
    std = np.abs(mean) * np.sqrt((Astd/Amean)**2 + (Bstd/Bmean)**2)
    return _xarray_werrmath_postprocess(A, B, mean, std)

@pcAccessor.register('werrdiv')
@format_docstring(**_paramdocs_werrmath)
def xarray_werrdiv(A, B, *, require_std=True):
    '''returns dataset of 'mean' and 'std' for A / B, including error propagation.
    Assumes independent errors and applies the "standard" formula:
        z = A / B
        mean(z) = mean(A) / mean(B)
        std(z) = abs(mean(z)) * sqrt((std(A)/mean(A))**2 + (std(B)/mean(B))**2)

    A, B: {A_B}
    require_std: {require_std}
    '''
    Amean, Astd, Bmean, Bstd = _xarray_werrmath_inputs(A, B, require_std=require_std)
    mean = Amean / Bmean
    std = np.abs(mean) * np.sqrt((Astd/Amean)**2 + (Bstd/Bmean)**2)
    return _xarray_werrmath_postprocess(A, B, mean, std)



"""
File Purpose: units in Bifrost
"""
from .bifrost_io_tools import read_bifrost_snap_idl
from ...mhd import MhdUnitsManager
from ...tools import format_docstring
from ...units import _units_manager_helpstr


''' --------------------- BifrostUnitsManager --------------------- '''

@format_docstring(helpstr=_units_manager_helpstr, sub_ntab=1)
class BifrostUnitsManager(MhdUnitsManager):
    '''units manager with the from_bifrost method,
    which determines all units based on only u_l, u_t, and u_r.

    Note: cgs units for electromagnetic quantities are not supported here,
        because cgs electromagnetic equations differ from SI equations,
        e.g. cgs includes extra factors of 4*pi or c in various places.

    How to infer electromagnetic units from t, l, and r alone?
        It is possible because 'raw' Bifrost units guarantee mu0 = 1,
            while in 'si', mu0 = 4 * pi * 10^-7 Newtons Ampere^-2
        Internally, this method determines:
            u_M = u_r * u_l**3   # mass
            u_N = u_M * u_l * u_t**-2   # energy (Newtons in SI)
        Amperes units:
            (mu0 [si]) = u_N * u_A**-2 * (mu0 [raw])
            --> u_A = sqrt(u_N * (mu0 [raw]) / (mu0 [si]))
            --> u_A = sqrt(u_N * (1) / (4 * pi * 10^-7))
        Which yields, for charge units:
            u_q = u_A * u_t
        At which point, we have u_l, u_t, u_M, and u_q,
            which is sufficient to infer all other units (aside from K for temperature)

    --- BifrostUnitsManager.help() will print a helpful message: ---
        {helpstr}
    '''
    @classmethod
    def from_bifrost_calculator(cls, bifrost_calculator, units='si', *, K=1):
        '''create a BifrostUnitsManager from a BifrostCalculator instance.

        bifrost_calculator: BifrostCalculator
            determine units based on this calculator's params: u_l, u_t, and u_r.
            (Assumes all snaps have the same u_l, u_t, and u_r.)
            CAUTION: the names u_l, u_t, u_r refer to cgs in Bifrost, but SI in PlasmaCalcs.
        units: 'si' or 'raw'
            by default, outputs of the resulting UnitsManager convert to this unit system.
            Can easily change later by setting result.units to a different value.
        K: number, default 1
            temperature [si] = K * temperature [raw]

        see help(cls) for more details on how the units are inferred.
        '''
        params = bifrost_calculator.params
        ucgs = dict(ucgs_l=params['u_l'],
                    ucgs_t=params['u_t'],
                    ucgs_r=params['u_r'])
        return cls.from_mhd_cgs(units=units, **ucgs, mu0_raw=1, K=K)

    @classmethod
    def from_snap_idl(cls, filename, units='si', *, K=1):
        '''create a BifrostUnitsManager from a Bifrost snap's snapname_NNN.idl file.

        filename: str
            path to the snap's IDL file.
        units: 'si' or 'raw'
            by default, outputs of the resulting UnitsManager convert to this unit system.
            Can easily change later by setting result.units to a different value.
        K: number, default 1
            temperature [si] = K * temperature [raw]

        see help(cls) for more details on how the units are inferred.
        '''
        params = read_bifrost_snap_idl(filename, eval=True)
        ucgs = dict(ucgs_l=params['u_l'],
                    ucgs_t=params['u_t'],
                    ucgs_r=params['u_r'])
        return cls.from_mhd_cgs(units=units, **ucgs, mu0_raw=1, K=K)



import PlasmaCalcs as pc
import matplotlib.pyplot as plt


dirs = {
    'EPPIC-2a': 'tfbi2a_prod1_n400',
    #'EPPIC-2a_n900': 'tfbi2a_prod1_n900',
    'EPPIC-2b': 'tfbi2_c3prod1_v2',
    '1 nscale, 4 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff08_A_fw=9_me4_nu4',
    '1 nscale, 16 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff09_nscale01_me16',
    '1 nscale, 64 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff09_nscale01_me64',
    '4 nscale, 1 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff07_A_fw=9_nscale004_nx=256',
    '4 nscale, 4 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff09_nscale04_me04',
    '4 nscale, 16 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff09_nscale04_me16',
    '4 nscale, 64 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff09_nscale04_me64',
    '16 nscale, 1 $m_e$': 'tfbi2_eff_RUNS/tfbi2_eff07_A_fw=9_nscale016_nx=128',
}

default_styles = dict(linewidth=2, alpha=0.7)
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
nscale_styles = {  # vary color for each nscale
    1: dict(color=colors[0]),
    4: dict(color=colors[1]),
    16: dict(color=colors[2]),
}
linestyles = ['-', '--', '-.', ':']
me_styles = {  # vary linestyle for each me
    1: dict(linestyle=linestyles[0]),
    4: dict(linestyle=linestyles[1]),
    16: dict(linestyle=linestyles[2]),
    64: dict(linestyle=linestyles[3]),
}

styles = {
    'EPPIC-2a': dict(color='black', linestyle='-', linewidth=4),
    'EPPIC-2b': dict(color='red', linestyle='-', linewidth=4, alpha=0.7),
    '1 nscale, 4 $m_e$': dict(**nscale_styles[1], **me_styles[4], **default_styles),
    '1 nscale, 16 $m_e$': dict(**nscale_styles[1], **me_styles[16], **default_styles),
    '1 nscale, 64 $m_e$': dict(**nscale_styles[1], **me_styles[64], **default_styles),
    '4 nscale, 1 $m_e$': dict(**nscale_styles[4], **me_styles[1], **default_styles),
    '4 nscale, 4 $m_e$': dict(**nscale_styles[4], **me_styles[4], **default_styles),
    '4 nscale, 16 $m_e$': dict(**nscale_styles[4], **me_styles[16], **default_styles),
    '4 nscale, 64 $m_e$': dict(**nscale_styles[4], **me_styles[64], **default_styles),
    '16 nscale, 1 $m_e$': dict(**nscale_styles[16], **me_styles[1], **default_styles),
}


calcs = dict()
for key, d in dirs.items():
    e_chdir(os.path.join('runs', d))
    ec = pc.EppicCalculator.from_here(snaps_from='timers', kw_units=dict(M=1))
    calcs[key] = ec

for key, cc in calcs.items():
    cc('rmscomps_Ta_from_moment2', component=[0,1,2], fluid=0, snap=cc.snaps.select_between(0, 0.015)).pc.timelines(label=key)



"""
File Purpose: tools related to plotting images or 3D data (e.g. imshow, pcolormesh)
"""
import xarray as xr
import matplotlib.pyplot as plt

from .labels import xarray_title_plot_node
from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings
from .plot_tools import (
    update_vlims, set_margins_if_provided, set_lims_if_provided, set_min_n_ticks,
    PlotDimsMixin,
    current_axes_or_None, current_figure_exists,
) 
from .scatter import (
    xarray_scatter_max, xarray_scatter_min, xarray_scatter_where,
)
from ..errors import (
    InputError,
    PlottingAmbiguityError,
)
from ..tools import (
    alias_key_of, alias_child, simple_property,
    UNSET,
    pcAccessor,
    xarray_ensure_dims, xarray_as_array, xarray_fill_coords,
)



''' --------------------- ImagePlotElement & Image --------------------- '''

_paramdocs_ax = {
    'ax': '''None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.''',
}

@PlotSettings.format_docstring(**_paramdocs_ax)
class ImagePlotElement(MoviePlotElement):
    '''image on an Axes.
    "image" refers to a matplotlib.cm.ScalarMappable object, e.g. the result of imshow or pcolormesh.

    array: array-like object, probably ndim=2.
        the data to be plotted. will be stored in self.data['array']
    ax: {ax}
    image_mode: {image_mode}

    Additional kwargs can be any PlotSettings; see help(self.plot_settings) for details.
    '''
    def __init__(self, array, *, ax=None,
                 image_mode=PlotSettings.default('image_mode'),  # <- could go in kw_super, but explicit is nice.
                  **kw_super):
        super().__init__(image_mode=image_mode, **kw_super)
        self._ax_init = ax
        self.array = array
        self.init_im()

    array = alias_key_of('data', 'array', doc='''array for plot. Internally, stored at self.data['array']''')
    im = simple_property('_im', doc='''the image object (instance of matplotlib.cm.ScalarMappable)''')
    ax = alias_child('im', 'axes', if_no_child=None, doc='''the axes containing this image''')
    fig = alias_child('ax', 'figure', if_no_child=None, doc='''figure containing this image''')

    # # # PLOTTING THIS ELEMENT # # #
    def init_im(self):
        '''initialize image; actually plot the data.
        stores plotted object in self.im & returns self.im.
        '''
        # plot settings / kwargs / bookkeeping
        image_mode = self.plot_settings['image_mode']
        settings_key_lookup = {'pcolormesh': 'plt.pcolormesh', 'imshow': 'plt.imshow'}
        if image_mode not in settings_key_lookup:
            raise InputError(f'invalid image_mode: {image_mode!r}, expected "pcolormesh" or "imshow".')
        settings_key = settings_key_lookup[image_mode]
        kw_plot = self.plot_settings.get_mpl_kwargs(settings_key)
        kw_plot = update_vlims(self.array, kw_plot)
        # ax to use for plot
        ax = self._ax_init
        if ax is None:
            ax = plt.gca()
        # make plot
        if image_mode == 'pcolormesh':
            self.im = ax.pcolormesh(self.array, **kw_plot)
        else:  # image_mode == 'imshow'
            self.im = ax.imshow(self.array, **kw_plot)
        return self.im

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''update the plot using data['array'].
        return list of all updated matplotlib Artist objects.
        '''
        array = data['array']
        if self.plot_settings['image_mode'] == 'pcolormesh':
            self.im.set_array(array)
        else:  # image_mode == 'imshow'
            self.im.set_data(array)
        self.array = array  # <- updated for bookkeeping/consistency.
        return [self.im]

    # # # DISPLAY # # #
    def __repr__(self):
        array_info = f'{type(self.array).__name__} with shape={self.array.shape}'
        return f'{type(self).__name__}({array_info})'


@PlotSettings.format_docstring(**_paramdocs_ax)
class Image(MoviePlotNode):
    '''stores an ImagePlotElement & has methods for plotting & updating the image!
    "image" refers to a matplotlib.cm.ScalarMappable object, e.g. the result of imshow or pcolormesh.

    array: array-like object, probably ndim=3.
        the data to be plotted. will be stored in self.data['array']
    '''
    def __init__(self, *args__TODO, **kw__TODO):
        raise NotImplementedError('[TODO]: MovieImage. In the meantime, use XarrayImage.')
        # no need to implement this if willing to convert arrays to xarrays before plotting.
        # most of PlasmaCalcs works with xarrays, so code development should focus on XarrayImage.


''' --------------------- XarrayImagePlotElement & XarrayImage --------------------- '''

@PlotSettings.format_docstring(**_paramdocs_ax)
class XarrayImagePlotElement(ImagePlotElement):
    '''image on an Axes, for an xarray.DataArray.
    "image" refers to a matplotlib.cm.ScalarMappable object, e.g. the result of imshow or pcolormesh.

    array: xarray.DataArray, probably ndim=2.
        the data to be plotted.
    ax: {ax}
    image_mode: {image_mode}

    add_colorbar: {add_colorbar}
    add_labels: {add_labels}

    min_n_ticks: {min_n_ticks}
    min_n_ticks_cbar: {min_n_ticks_cbar}

    grid: {grid}
    '''
    def __init__(self, array, *, ax=None,
                 image_mode=PlotSettings.default('image_mode'),  # <- could go in kw_super, but explicit is nice.
                 add_colorbar=PlotSettings.default('add_colorbar'),
                 add_labels=PlotSettings.default('add_labels'),
                 **kw_super):
        kw_super.update(image_mode=image_mode, add_colorbar=add_colorbar, add_labels=add_labels)
        super().__init__(array, ax=ax, **kw_super)
    
    @property
    def cbar(self):
        '''Colorbar object associated with this image. False if no associated colorbar.
        None if self.im does not exist yet (i.e. not yet plotted).
        "associated" meaning: colorbar using this image as its scalar mappable.
        '''
        im = getattr(self, 'im', None)
        if im is None:
            return None
        else:
            result = getattr(im, 'colorbar', None)
            if result is None:  # im exists but has no colorbar
                return False
            else:
                return result

    def init_im(self):
        '''initialize image on self.ax; actually plot the data.
        stores plotted object in self.im & returns self.im.
        '''
        array = self.array
        image_mode = self.plot_settings['image_mode']
        kw_plot = self.plot_settings.get_mpl_kwargs('xarray.DataArray.plot[image]')
        kw_plot = update_vlims(array, kw_plot)
        plotter = getattr(array.plot, image_mode)  # e.g. array.plot.pcolormesh
        ax = self._ax_init
        if ax is None:
            ax = current_axes_or_None()
            if (ax is None) and (not current_figure_exists()):  # no current axes nor figure.
                kw_fig = self.plot_settings.get_mpl_kwargs('plt.figure')
                _fig = plt.figure(**kw_fig)
                # ax = None, still, but plotter can now create ax on current fig as needed.
            if (ax is None):
                polar = self.plot_settings.get('polar', last_resort_default=False)
                if polar:
                    ax = plt.gcf().add_subplot(polar=polar)
        self.im = plotter(ax=ax, **kw_plot)
        # labels
        if self.plot_settings['add_labels']:
            self.plot_labels()
        # formatting
        aspect = self.plot_settings['aspect']  # xarray plotter doesn't like 'aspect' kwarg; assign here instead.
        if aspect is not None:
            self.ax.set_aspect(aspect)
        set_margins_if_provided(self.plot_settings, ax=self.ax)
        set_lims_if_provided(self.plot_settings, ax=self.ax)
        min_n_ticks = self.plot_settings['min_n_ticks']
        if min_n_ticks is not None:
            fail_ok=self.plot_settings['min_n_ticks_fail_ok']
            set_min_n_ticks(min_n_ticks, ax=self.ax, fail_ok=fail_ok)
        if self.cbar is not False:
            min_n_ticks_cbar = self.plot_settings['min_n_ticks_cbar']
            if min_n_ticks_cbar is not None:
                fail_ok = self.plot_settings['min_n_ticks_cbar_fail_ok']
                set_min_n_ticks(min_n_ticks_cbar, ax=self.cbar.ax, fail_ok=fail_ok)
        grid = self.plot_settings.get('grid')
        if grid is not UNSET:
            if isinstance(grid, dict):
                self.ax.grid(**grid)
            else:
                self.ax.grid(grid)
        return self.im

    def plot_labels(self):
        '''plots xlabel and ylabel from self.plot_settings.
        If xlabel or ylabel not provided:
            - use defaults from xarray's plot method.
            - if array.attrs has 'coords_units', put that in brackets, e.g. 'x [si]'.
        If they are provided, use them, but also .format(**array.attrs).
            (e.g. 'x [{units}]' becomes 'x [si]' if array.attrs['units'] == 'si')
        '''
        xlabel = self.plot_settings.get('xlabel', default=None)
        ylabel = self.plot_settings.get('ylabel', default=None)
        if xlabel is None:
            if 'coords_units' in self.array.attrs:
                coords_units = self.array.attrs['coords_units']
                xlabel = ' '.join((self.ax.get_xlabel(), f'[{coords_units}]'))
                self.ax.set_xlabel(xlabel)
        else:
            xlabel = xlabel.format(**self.array.attrs)
            self.ax.set_xlabel(xlabel)
        if ylabel is None:
            if 'coords_units' in self.array.attrs:
                coords_units = self.array.attrs['coords_units']
                ylabel = ' '.join((self.ax.get_ylabel(), f'[{coords_units}]'))
                self.ax.set_ylabel(ylabel)
        else:
            ylabel = ylabel.format(**self.array.attrs)
            self.ax.set_ylabel(ylabel)


@pcAccessor.register('image')
@PlotSettings.format_docstring(**_paramdocs_ax)
class XarrayImage(MoviePlotNode, PlotDimsMixin):
    '''MoviePlotNode of an xarray.DataArray.
    stores an XarrayImagePlotElement & has methods for plotting & updating the image!
    "image" refers to a matplotlib.cm.ScalarMappable object, e.g. the result of imshow or pcolormesh.

    array: xarray.DataArray, probably ndim=3; or xarray.Dataset with single data_var
        the data to be plotted. if ndim=2, can still plot, but nothing to animate.
        if xarray.Dataset, must have only one data_var; will create image of that data_var.
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None -> infer from array & any other provided dimensions.
        if provided but dimension not found, attempt xarray_promote_dim before crashing.
    x, y: None or str
        if provided, tells dimensions for the x, y plot axes.
        None -> infer from array & any other provided dimensions.
        if provided but dimension not found, attempt xarray_promote_dim before crashing.
    ax: {ax}
    image_mode: {image_mode}
    init_plot: {init_plot}

    {kw_ax_margin}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}

    grid: {grid}

    # [TODO] example
    '''
    element_cls = XarrayImagePlotElement

    def __init__(self, array, t=None, *, x=None, y=None, ax=None,
                 image_mode=PlotSettings.default('image_mode'),  # <- could go in kw_super, but explicit is nice.
                 init_plot=PlotSettings.default('init_plot'),
                 add_colorbar=PlotSettings.default('add_colorbar'),
                 add_labels=PlotSettings.default('add_labels'),
                 title=PlotSettings.default('title'),
                 **kw_super):
        if isinstance(array, xr.Dataset):
            array = xarray_as_array(array)
        array = xarray_fill_coords(array, need=[x, y, t])
        array = xarray_ensure_dims(array, ({x, y, t} - {None}), promote_dims_if_needed=True)
        self.array = array
        self.t = t
        self.x = x
        self.y = y
        self._ax_init = ax
        kw_super.update(image_mode=image_mode, init_plot=init_plot,
                        add_colorbar=add_colorbar, add_labels=add_labels,
                        title=title,
                        )
        self.plot_dims_attrs = self.plot_dims_attrs.copy()  # copy cls attr to avoid overwriting by accident
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    ax = alias_child('obj', 'ax', if_special_child={None: None, UNSET: None},
        doc='''mpl.axes.Axes where this XarrayImage is plotted, or None if not plotted.''')

    fig = alias_child('obj', 'fig', if_special_child={None: None, UNSET: None},
        doc='''figure where this XarrayImage is plotted, or None if not plotted.''')

    im = alias_child('obj', 'im', if_special_child={None: None, UNSET: None},
        doc='''mpl.cm.ScalarMappable object of this XarrayImage, or None if not plotted.''')

    cbar = alias_child('obj', 'cbar', if_special_child={None: None, UNSET: None},
        doc='''the mpl.colorbar.Colorbar of this XarrayImage.
        None if image not plotted.
        False if image plotted but does not have an associated colorbar.''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the XarrayImagePlotElement at self.obj.'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs('pc.XarrayImagePlotElement')
        # -- determine vlims (based on the entire array, not just this frame)
        kw_plot = update_vlims(self.array, kw_plot)
        # >> actually plot the image <<
        self.obj = self.element_cls(data['array'], ax=self._ax_init, **kw_plot)
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        # bookkeeping
        self.frame = frame

    def get_data_at_frame(self, frame):
        '''returns {'array': array at this frame}, properly transposed & ready for plotting.'''
        t = self.t_plot_dim
        result = self.array
        if t is not None:  # there is a dimension for the time axis.
            result = result.isel({t: frame})
        x, y = self.xy_plot_dims   # (x and y are strings, the dimension names)
        result = result.transpose(y, x)   # transposed to get correct order for plotting
        return {'array': result}

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t_plot_dim
        if t is None:
            return 1
        return len(self.array.coords[t])

    # # # INFERRING DIMS (REQUIRED BY PARENT CLASS) # # #
    plot_dims_attrs = dict(x='x', y='y', t='t', dims='_array_dims', t_necessary_if='_array_is_not_2D')
    _array_dims = property(lambda self: self.array.dims, doc='''the dimensions of the array''')
    # x_plot_dim, y_plot_dim, t_plot_dim are already defined by PlotDimsMixin.

    def _array_is_not_2D(self):
        '''tells whether self.array is not 2D'''
        return self.array.ndim != 2

    def _has_valid_plot_dims(self):
        '''return whether this array is plottable, i.e. all plot_dims can be inferred,
        and array doesn't have more than 3 dims.'''
        if self.array.ndim > 3:
            return False
        try:
            _plot_dims = self.xyt_plot_dims
        except PlottingAmbiguityError:
            return False
        return True

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is None:
            self.ax.set_title('')  # remove xarray's default title.
        else:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.array, title, t=self.t_plot_dim,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging

    # # # MISC CONVENIENT METHODS # # #
    def scatter_max(self, **kw_scatter):
        '''use XarrayScatter to plt.scatter() a single marker, at the argmax of self.array.
        animatable (e.g. different max for each frame); does self.add_child(scatter result)
        default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MAX}
        returns XarrayScatter object (which got added as child of self).

        see also: xarray_scatter_max (or, array.pc.scatter_max)
        '''
        kw_scatter = {'add_labels': False,   # by default, don't let scatter overwrite image title.
                      **kw_scatter}
        result = xarray_scatter_max(self.array, x=self.x_plot_dim, y=self.y_plot_dim,
                                    t=self.t_plot_dim, ax=self.ax, **kw_scatter)
        self.add_child(result)
        return result

    def scatter_min(self, **kw_scatter):
        '''use XarrayScatter to plt.scatter() a single marker, at the argmin of self.array.
        animatable (e.g. different min for each frame); does self.add_child(scatter result)
        default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MIN}
        returns XarrayScatter object (which got added as child of self).

        see also: xarray_scatter_min (or, array.pc.scatter_min)
        '''
        kw_scatter = {'add_labels': False,   # by default, don't let scatter overwrite image title.
                      **kw_scatter}
        result = xarray_scatter_min(self.array, x=self.x_plot_dim, y=self.y_plot_dim,
                                    t=self.t_plot_dim, ax=self.ax, **kw_scatter)
        self.add_child(result)
        return result

    def scatter_where(self, condition, **kw_scatter):
        '''use XarrayScatter to plt.scatter() markers where condition is True.
        animatable (e.g. different condition for each frame); does self.add_child(scatter result)
        returns XarrayScatter object (which got added as child of self).

        condition: callable or DataArray of bools
            callable --> use condition(self.array)
            DataArray --> ignore self.array entirely.

        For x, y, t plot dims, uses self.x_plot_dim, self.y_plot_dim, self.t_plot_dim.

        see also: xarray_scatter_where (or, array.pc.scatter_where)

        Example:
            xim = array.pc.image(...)
            xim.scatter_where(lambda arr: arr > 7.5)  # marker at all points where arr > 7.5
            xim.save(filename)  # save animation to filename (if xim.t_plot_dim is not None)
        '''
        kw_scatter = {'add_labels': False,   # by default, don't let scatter overwrite image title.
                      **kw_scatter}
        if callable(condition):
            condition = condition(self.array)
        x, y, t = self.x_plot_dim, self.y_plot_dim, self.t_plot_dim
        condition = xarray_fill_coords(condition, need=[x,y,t])
        condition = xarray_ensure_dims(condition, ({x, y, t} - {None}), promote_dims_if_needed=True)
        # [TODO] ^ be more lenient if appropriate coords don't exist in condition?
        result = xarray_scatter_where(condition, x=x, y=y, t=t, ax=self.ax, **kw_scatter)
        self.add_child(result)
        return result



"""
File Purpose: tools for addons
"""

ADDON_LOADERS = []  # list of QuantityLoader addon classes
ADDONS_FAILED = dict()  # {name: ImportFailed object or list of ImportFailed objects}

def register_addon_loader(loader):
    '''registers loader as an existing addon loader. Then returns loader.'''
    ADDON_LOADERS.append(loader)
    return loader

def register_addon_loader_if(condition):
    '''returns registers_addon_loader function if condition, else function that does nothing.'''
    return register_addon_loader if condition else lambda x: x

def register_addon_failed(name):
    '''registers name as a failed addon.'''
    ADDONS_FAILED.append(name)



"""
Package Purpose: PlasmaCalculator for Bifrost
"""
# # # SINGLE FLUID STUFF # # #
from .bifrost_bases import BifrostBasesLoader
from .bifrost_calculator import BifrostCalculator
from .bifrost_snaps import BifrostSnap, BifrostSnapList, BifrostScrSnap
from .bifrost_direct_loader import BifrostDirectLoader
from .bifrost_efield import BifrostEfieldLoader
from .bifrost_io_tools import (
    read_bifrost_snap_idl,
    bifrost_snap_idl_files,
    bifrost_infer_snapname_here,
    read_bifrost_meshfile, slice_bifrost_meshfile,
    BifrostVarPathsManager,
    BifrostDataCutter,
)
from .bifrost_stagger import (
    StaggerConstants,
    STAGGER_ABC_DERIV_o1, STAGGER_ABC_SHIFT_o1,
    STAGGER_ABC_DERIV_o5, STAGGER_ABC_SHIFT_o5,
    transpose_to_0, transpose_to_0_tuple, simple_slice,
    Staggerer, StaggerInterface3D,
    BifrostStaggerable,
)
from .bifrost_units import BifrostUnitsManager

# # # MULTIFLUID STUFF # # #
from .bifrost_multifluid import (
    BifrostMultifluidDensityLoader,
    BifrostMultifluidCalculator,
)
